"""
This is a script for the frontend of the ShapExplainer class used to obtain Mote Carlo estimates of feature impact.
"""
from typing import Optional, Dict, List, Tuple, Type, Union, Callable
import pandas as pd
import numpy as np
from sklearn.base import BaseEstimator
from sklearn.ensemble import RandomForestClassifier, ExtraTreesClassifier
import warnings
import shap

import aaanalysis.utils as ut
from .backend.check_models import (check_match_list_model_classes_kwargs,
                                   check_match_labels_X,
                                   check_match_X_is_selected)
from .backend.shap_explainer.shap_explainer_fit import monte_carlo_shap_estimation
from .backend.shap_explainer.se_add_feat_impact import (comp_shap_feature_importance,
                                                        insert_shap_feature_importance,
                                                        comp_shap_feature_impact,
                                                        insert_shap_feature_impact)
from .backend.shap_explainer.se_add_sample_mean_dif import add_sample_mean_dif_


# I Helper Functions
# Check init
def check_shap_explainer(explainer_class=None, explainer_kwargs=None):
    """Check if explainer class is a valid shap explainer"""
    ut.check_mode_class(model_class=explainer_class)
    ut.check_dict(name="explainer_kwargs", val=explainer_kwargs, accept_none=True)
    list_valid_explainers = [shap.TreeExplainer, shap.LinearExplainer, shap.KernelExplainer,
                             shap.DeepExplainer, shap.GradientExplainer]
    names_valid_explainers = [x.__name__ for x in list_valid_explainers]
    if not (isinstance(explainer_class, type) and explainer_class in list_valid_explainers):
        raise ValueError(f"'{explainer_class.__name__}' is not a valid 'explainer_class'. "
                         f"Chose from the following: {names_valid_explainers}")
    # Check if explainer has shap_values method
    explainer_kwargs = ut.check_model_kwargs(model_class=explainer_class,
                                             model_kwargs=explainer_kwargs,
                                             name_model_class="explainer_class",
                                             method_to_check="shap_values")
    return explainer_kwargs


def check_match_class_explainer_and_models(explainer_class=None, explainer_kwargs=None, list_model_classes=None):
    """Check if each model in list_model_class is compatible with the shap explainer class"""
    dummy_data = np.array([[0, 1], [1, 0]])  # Minimal dummy data to initialize explainers
    dummy_label = [0, 1]
    for model_class in list_model_classes:
        # Check model compatability
        try:
            # Fit the dummy model
            model = model_class().fit(dummy_data, dummy_label)
            # Determine the correct input for the explainer
            if explainer_class.__name__ in ['KernelExplainer', 'OtherExplainerNeedingPredict']:
                model_input = model.predict_proba if hasattr(model, 'predict_proba') else model.predict
            else:
                model_input = model
            # Attempt to create the explainer with the appropriate input
            explainer = explainer_class(model_input, dummy_data, **{})
        except Exception as e:
            str_error = (f"The SHAP explainer '{explainer_class.__name__}' is not compatible with "
                         f"the model '{model_class.__name__}'.\nSHAP message:\n\t{e}")
            raise ValueError(str_error)
        if explainer_kwargs is not None:
            try:
                explainer = explainer_class(model_input, dummy_data, **explainer_kwargs)
            except Exception as e:
                str_error = (f"The SHAP explainer '{explainer_class.__name__}' has invalid 'explainer_kwargs': {explainer_class}"
                             f"\nSHAP message:\n\t{e}")
                raise ValueError(str_error)


# Check functions for fit method
def check_shap_values(shap_values=None):
    """Check if shap values are properly set"""
    if shap_values is None:
        raise ValueError("'shape_values' are None. Use 'ShapExplainer().fit()' to compute them.")
    ut.check_array_like(name="shap_values", val=shap_values, dtype="numeric")


def check_match_labels_X_fuzzy_labeling(labels=None, X=None, fuzzy_labeling=False):
    """Check if labels are binary classification task labels or apply to fuzzy_labeling [0-1]"""
    if not fuzzy_labeling:
        labels = check_match_labels_X(labels=labels, X=X)
        return labels
    n_samples = X.shape[0]
    # Accept float if fuzzy_labeling is True
    labels = ut.check_labels(labels=labels, len_requiered=n_samples, accept_float=fuzzy_labeling)
    unique_labels = set(labels)
    if len(unique_labels) < 2:
        raise ValueError(f"'labels' should contain at least 2 unique labels ({unique_labels})")
    # The higher label is considered as the positive (test) class
    wrong_labels = [x for x in labels if not 0 <= x <= 1]
    if len(wrong_labels) != 0:
        raise ValueError(f"'labels' should contain only values between 0 and 1. Wrong labels are: {wrong_labels}")
    return labels


def check_is_selected(is_selected=None, n_feat=None):
    """Check is_selected and set if None"""

    if is_selected is None:
        is_selected = np.ones((1, n_feat), dtype=bool)
    else:
        is_selected = ut.check_array_like(name="is_selected_feature", val=is_selected, accept_none=False)
        is_selected = is_selected.astype(bool)
        is_selected = ut.check_array_like(name="is_selected_feature", val=is_selected, accept_none=False,
                                          expected_dim=2, dtype="bool")
    return is_selected


def check_match_labels_fuzzy_labeling(labels=None, fuzzy_labeling=False, verbose=True):
    """Check if only on label is fuzzy labeled and that the remaining sample balanced (best training scenario)"""
    # TODO adjust for multi-class
    if not fuzzy_labeling:
        return  # Skip check if fuzzy labeling is not enabled
    # Check for one fuzzy label and balance among other labels
    n_fuzzy_labels = len([label for label in labels if label not in [0, 1]])
    if n_fuzzy_labels != 1 and verbose:
        warnings.warn(f"Optimal training with fuzzy labeling requires exactly one fuzzy label, but {n_fuzzy_labels} were given.")
    non_fuzzy_labels = [label for label in labels if label in [0, 1]]
    n_pos = non_fuzzy_labels.count(1)
    n_neg = non_fuzzy_labels.count(0)
    if n_pos != n_neg and verbose:
        warnings.warn(f"For optimal training, non-fuzzy labels should be balanced (equal number of 0s and 1s). "
                      f"\n The 'labels' contain {n_pos} positive (1) and {n_neg} negative (0) samples.")


def check_match_labels_target_class_labels(label_target_class=None, labels=None):
    """Check if class index matches to classes (unique labels) in labels"""
    label_classes = sorted(list(dict.fromkeys([x for x in labels if x == int(x)])))
    if label_target_class not in label_classes:
        raise ValueError(f"'label_target_class' ({label_target_class}) should be from 'labels' classes: {label_classes}")


def check_match_n_background_data_X(n_background_data=None, X=None):
    """Check if n_background_data not >= n_samples"""
    if n_background_data is None:
        return None # Skip test
    n_samples, n_features = X.shape
    if n_background_data >= n_samples:
        raise ValueError(f"'n_background_data' ({n_background_data}) must be < 'n_samples' ({n_samples}) from 'X'.")


# Check function for add_feat_impact method
def check_sample_positions(sample_positions=None, n_samples=None):
    """Check if sample_positions is int or list of ints"""
    str_add = "'sample_position' should be integer or list of integers indicating sample indices."
    if sample_positions is None:
        # Create sample_positions list for all samples
        sample_positions = list(range(n_samples))
    args = dict(min_val=0, max_val=n_samples-1, just_int=True, str_add=str_add)
    if isinstance(sample_positions, list) or isinstance(sample_positions, np.ndarray):
        ut.check_list_like(name="sample_positions", val=sample_positions, check_all_non_neg_int=True, str_add=str_add)
        for i in sample_positions:
            ut.check_number_range(name=f"sample_positions: {i}", val=i, **args)
    else:
        ut.check_number_range(name=f"sample_positions", val=sample_positions, **args)
    return sample_positions


def check_names(names=None):
    """Check if name is str ror list of str"""
    if names is None:
        return None     # Skip test
    str_add = f"'name' should be string or list of strings, but following was given: {names}"
    if isinstance(names, list):
        for i in names:
            ut.check_str(name=f"names: {i}", val=i, str_add=str_add)
        duplicated_names = list(set([x for x in names if names.count(x) > 1]))
        if len(duplicated_names) > 0:
            raise ValueError(f"'names' should not contain duplicated names: {duplicated_names}")
    else:
        ut.check_str(name=f"names", val=names, str_add=str_add)


def check_match_sample_positions_names(sample_positions=None, names=None, group_average=False):
    """Check if length of sample_positions and names matches"""
    # Group scenario
    if group_average:
        if not isinstance(sample_positions, list):
            raise ValueError(f"For 'group_average', 'sample_positions' ({sample_positions}) must be a list of integers.")
        if names is None:
            names = "Group"  # Default group names
        elif isinstance(names, str):
            names = names  # Ensure names is a list
        else:
            raise ValueError(f"For 'group_average', 'names' ({names}) must be a single string or None.")
    # Single sample or multiple samples scenarios
    else:
        if names is not None:
            if isinstance(sample_positions, int) and isinstance(names, str):
                sample_positions = [sample_positions]  # Convert to list for consistency
                names = [names]
            elif isinstance(sample_positions, list) and isinstance(names, list):
                if len(sample_positions) != len(names):
                    raise ValueError(f"Length of 'sample_positions' (n={len(sample_positions)}) and 'names' (n={len(names)}) must be equal for multiple samples.")
            else:
                raise ValueError("Mismatch: 'sample_positions' should be int and 'names' str for a single sample, and both lists for multiple samples."
                                 f"\n 'sample_positions': {sample_positions}. \n 'names': {names}")
        # Generate default names if names is None
        else:
            if isinstance(sample_positions, list):
                names = [f"Protein{p}" for p in sample_positions]
            else:
                names = [f"Protein{sample_positions}"]
                sample_positions = [sample_positions]
    return sample_positions, names


def check_match_sample_positions_group_average(sample_positions=None, group_average=False):
    """Check if group_average only True if sample_positions is list"""
    if group_average:
        ut.check_list_like(name="sample_positions", val=sample_positions)
        if len(sample_positions) == 1:
            raise ValueError


def check_match_df_feat_shap_values(df_feat=None, shap_values=None, drop=False, shap_feat_importance=False):
    """Check if df_feat matches with importance values"""
    if shap_values is None:
        raise ValueError(f"'shap_values' is None. Please fit ShapExplainer before adding feature impact.")
    n_feat = len(df_feat)
    n_samples, n_feat_imp = shap_values.shape
    if n_feat != n_feat_imp:
        raise ValueError(f"Mismatch of number of features in 'df_feat' (n={n_feat} and in 'shap_values' (n={n_feat_imp})")
    if not drop:
        # Check if columns already exist
        if shap_feat_importance:
            if ut.COL_FEAT_IMPORT in list(df_feat):
                raise ValueError(f"'{ut.COL_FEAT_IMPORT}' already in 'df_feat' columns. To override, set 'drop=True'.")
        else:
            existing_impact_cols = [x for x in list(df_feat) if ut.COL_FEAT_IMPACT in x]
            if len(existing_impact_cols) > 0:
                raise ValueError(f"Some '{ut.COL_FEAT_IMPACT}' columns exist already in 'df_feat'. "
                                 f"To override, set 'drop=True'.\n These columns comprise: {existing_impact_cols}")


# II Main Functions
class ShapExplainer:
    """
    SHAP Explainer class: A wrapper for SHAP (SHapley Additive exPlanations) explainers to obtain Monte Carlo estimates
    for feature impact [Breimann24c]_.

    `SHAP <https://shap.readthedocs.io/en/latest/index.html>`_ is an explainable Artificial Intelligence (AI) framework
    and game-theoretic approach to explain the output of any machine learning model using SHAP values. These SHAP values
    represent a feature's responsibility for a change in the model output to increase of decrease a sample prediction
    score due to the positive or negative impact of its features, respectively.

    Attributes
    ----------
    shap_values : array-like, shape(n_samples, n_features)
        2D array with Monte Carlo estimates of SHAP values obtained by SHAP explainer models averaged across all rounds,
        feature selections, and trained models from `list_model_classes`.
    exp_value : int
        Expected value for explaining the model output obtained by SHAP explainer model  averaged across all rounds,
        feature selections, and trained models from `list_model_classes`. Typically, 0.5 for binary classification
        and balanced dataset.

    """
    def __init__(self,
                 explainer_class: Callable = shap.TreeExplainer,
                 explainer_kwargs: Optional[dict] = None,
                 list_model_classes: List[Type[Union[BaseEstimator]]] = None,
                 list_model_kwargs: Optional[List[dict]] = None,
                 verbose: bool = True,
                 random_state: Optional[int] = None,
                 ):
        """
        Parameters
        ----------
        explainer_class : model, default=TreeExplainer
            The `SHAP Explainer model <https://shap.readthedocs.io/en/latest/api.html#explainers>`_.
            Must be one of the following: :class:`shap.TreeExplainer`, :class:`shap.LinearExplainer`,
            :class:`shap.KernelExplainer`, :class:`shap.DeepExplainer`, :class:`shap.GradientExplainer`.
        explainer_kwargs : dict, default={model_output='probability'}
            Keyword arguments for the explainer class model.
        list_model_classes : list of Type[BaseEstimator], default=[RandomForestClassifier, ExtraTreesClassifier]
            A list of prediction model classes used to obtain SHAP values.
        list_model_kwargs : list of dict, optional
            A list of dictionaries containing keyword arguments for each model in `list_model_classes`.
        verbose : bool, default=True
            If ``True``, verbose outputs are enabled.
        random_state : int, optional
            The seed used by the random number generator. If a positive integer, results of stochastic processes are
            consistent, enabling reproducibility. If ``None``, stochastic processes will be truly random.

        Notes
        -----
        * All attributes are set during fitting via the :meth:`ShapExplainer.fit` method and can be directly accessed.
        * The Explainer models should be provided from the `SHAP <https://shap.readthedocs.io/en/latest/index.html>`_
          package
        * SHAP model fitting messages appear in red and are not controlled by ``verbose``, unlike AAanalysis progress
          messages in blue.
        * The selection of the SHAP explainer must align with the machine learning models used.
          Following explainer model types are allowed:

          - :class:`shap.TreeExplainer`: Ideal for tree-based models (by default, random forests and extra trees; further
            recommended are XGBoost and CatBoost). Efficient in computing SHAP values by leveraging the tree structure.
          - :class:`shap.LinearExplainer`: Suited for linear models (e.g., logistic regression, linear regression).
            Computes SHAP values directly from model coefficients.
          - :class:`shap.KernelExplainer`: Model-agnostic, works with any model type. Uses weighted linear regression to
            approximate SHAP values. Versatile but less computationally efficient, which can be increased by a background dataset.
          - :class:`shap.DeepExplainer`: Designed for deep learning models (e.g., models from TensorFlow, Keras).
            Approximates SHAP values by analyzing neuron groups, suitable for complex networks.
          - :class:`shap.GradientExplainer`: Also for deep learning, but uses expected gradients.
            Effective for models with differentiable components.

          Proper explainer choice is key for accurate model explanations.

        * By default, :class:`shap.TreeExplainer` is used with random forest, extra trees, and gradient boosting models.

        See Also
        --------
        * :class:`sklearn.ensemble.RandomForestClassifier` for random forest model.
        * :class:`sklearn.ensemble.ExtraTreesClassifier` for extra trees model.
        * :meth:`ShapExplainer.add_feat_impact` for details on feature impact and SHAP value-based feature importance.

        Examples
        --------
        .. include:: examples/shap_explainer.rst
        """
        # Global parameters
        verbose = ut.check_verbose(verbose)
        random_state = ut.check_random_state(random_state=random_state)
        # Check SHAP explainer model
        if explainer_class is None:
            explainer_class = shap.TreeExplainer
            explainer_kwargs = explainer_kwargs or dict(model_output="probability")
        explainer_kwargs = check_shap_explainer(explainer_class=explainer_class, explainer_kwargs=explainer_kwargs)
        # Check model parameters
        if list_model_classes is None:
            list_model_classes = [RandomForestClassifier, ExtraTreesClassifier]
        elif not isinstance(list_model_classes, list):
            # If single model is used (not recommended)
            list_model_classes = [list_model_classes]
        list_model_classes = ut.check_list_like(name="list_model_classes",
                                                val=list_model_classes,
                                                accept_none=False,
                                                min_len=1)
        ut.check_list_like(name="list_model_kwargs", val=list_model_kwargs, accept_none=True)
        if list_model_kwargs is None:
            list_model_kwargs = [{} for _ in list_model_classes]
        # Check matching of model parameters
        check_match_list_model_classes_kwargs(list_model_classes=list_model_classes,
                                              list_model_kwargs=list_model_kwargs)
        _list_model_kwargs = []
        for model_class, model_kwargs in zip(list_model_classes, list_model_kwargs):
            ut.check_mode_class(model_class=model_class)
            model_kwargs = ut.check_model_kwargs(model_class=model_class,
                                                 model_kwargs=model_kwargs,
                                                 random_state=random_state)
            _list_model_kwargs.append(model_kwargs)
        check_match_class_explainer_and_models(explainer_class=explainer_class,
                                               explainer_kwargs=explainer_kwargs,
                                               list_model_classes=list_model_classes)
        # Internal attributes
        self._verbose = verbose
        self._random_state = random_state
        self._explainer_class = explainer_class
        self._explainer_kwargs = explainer_kwargs
        self._list_model_classes = list_model_classes
        self._list_model_kwargs = _list_model_kwargs
        # Output parameters (set during model fitting)
        self.shap_values: Optional[ut.ArrayLike2D] = None
        self.exp_value: Optional[int] = None

    def fit(self,
            X: ut.ArrayLike2D,
            labels: ut.ArrayLike1D = None,
            label_target_class: int = 1,
            n_rounds: int = 5,
            is_selected: ut.ArrayLike2D = None,
            fuzzy_labeling: bool = False,
            n_background_data: Optional[int] = None,
            ) -> "ShapExplainer":
        """
        Obtain SHAP values aggregated across prediction models and training rounds.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix. `Rows` typically correspond to proteins and `columns` to features.
        labels : array-like, shape (n_samples)
            Class labels for samples in ``X`` (typically, 1=positive, 0=negative).
        label_target_class : int, default=1
            The label of the class for which SHAP values are computed in a classification tasks.
            For binary classification, '0' represents the negative class and '1' the positive class.
        n_rounds : int, default=5
            The number of rounds (>=1) to fit the models and obtain the SHAP values by explainer.
        is_selected : array-like, shape (n_selection_round, n_features)
            2D boolean arrays indicating different feature selections.
        fuzzy_labeling : bool, default=False
            If ``True``, fuzzy labeling is applied to approximate SHAP values for samples with uncertain/partial
            memberships (e.g., between >0 and <1 for binary classification scenarios).
        n_background_data : None or int, optional
            The number samples (< 'n_samples') in the background dataset used for the `KernelExplainer`` to reduce
            computation time. The dataset is obtained by k-means clustering. If ``None``, the full dataset 'X' is used.

        Returns
        -------
        ShapExplainer
            The fitted ShapExplainer model instance.

        Notes
        -----
        **Fuzzy Labeling**

        * Aim: Compute SHAP value for datasets with uncertain or ambiguous labels. Especially useful to explain newly
          predicted samples, where class label is set to the respective prediction probability.
        * Approach: Uses probabilistic labels to represent degrees of membership.
        * Idea: Adjusts label thresholds dynamically in Monte Carlo estimation to better represent label uncertainties.
        * Background: Inspired by fuzzy logic, replacing binary true/false with degrees of truth.

        See Also
        --------
        * [Breimann24c]_ introduces fuzzy labeling to compute Monte Carlo estimates of SHAP values
          for samples with not clearly defined class membership.

        Examples
        --------
        .. include:: examples/se_fit.rst
        """
        # Check input
        X = ut.check_X(X=X)
        n_samples, n_feat = X.shape
        ut.check_X_unique_samples(X=X, min_n_unique_samples=2)
        ut.check_bool(name="fuzzy_labeling", val=fuzzy_labeling)
        labels = check_match_labels_X_fuzzy_labeling(labels=labels, X=X, fuzzy_labeling=fuzzy_labeling)
        check_match_labels_fuzzy_labeling(labels=labels, fuzzy_labeling=fuzzy_labeling, verbose=self._verbose)
        ut.check_number_range(name="label_target_class", val=label_target_class, min_val=0, just_int=True, accept_none=True)
        check_match_labels_target_class_labels(label_target_class=label_target_class, labels=labels)
        is_selected = check_is_selected(is_selected=is_selected, n_feat=n_feat)
        check_match_X_is_selected(X=X, is_selected=is_selected)
        ut.check_number_range(name="n_rounds", val=n_rounds, min_val=1, just_int=True)
        ut.check_number_range(name="n_background_data", val=n_background_data, min_val=1, just_int=True, accept_none=True)
        check_match_n_background_data_X(n_background_data=n_background_data, X=X)
        # Compute SHAP values
        shap_values, exp_val = monte_carlo_shap_estimation(X, labels=labels,
                                                           list_model_classes=self._list_model_classes,
                                                           list_model_kwargs=self._list_model_kwargs,
                                                           explainer_class=self._explainer_class,
                                                           explainer_kwargs=self._explainer_kwargs,
                                                           is_selected=is_selected,
                                                           fuzzy_labeling=fuzzy_labeling,
                                                           n_rounds=n_rounds,
                                                           verbose=self._verbose,
                                                           label_target_class=label_target_class,
                                                           n_background_data=n_background_data)
        self.shap_values = shap_values
        self.exp_value = exp_val
        return self

    def eval(self, shap_values=None, is_selected=None):
        """
        UNDER CONSTRUCTION - Evaluate convergence of Monte Carlo estimates of SHAP values depending on number of rounds
        """

    def add_feat_impact(self,
                        df_feat: pd.DataFrame = None,
                        drop: bool = False,
                        sample_positions: Union[int, List[int], None] = None,
                        names: Optional[Union[str, List[str]]] = None,
                        normalize: bool = True,
                        group_average: bool = False,
                        shap_feat_importance: bool = False,
                        ) -> pd.DataFrame:
        """
        Compute SHAP feature impact (or importance) from SHAP values and add to the feature DataFrame.

        Three different scenarios for computing the feature impact are possible:

            a) **Single sample**: Computes the feature impact for a selected sample.
            b) **Multiple samples**: Computes the feature impact for multiple samples (all by default).
            c) **Group of samples**: Computes the average feature impact for a group of samples (+ standard deviation).

        The calculated feature impacts are added to ``df_feat`` as new columns named ``feat_impact_'name(s)'``,
        corresponding to each sample or group. Additionally, the SHAP value-based feature importance can be included
        as ``feat_importance`` column.

        Parameters
        ----------
        df_feat : pd.DataFrame, shape (n_features, n_feature_info)
            Feature DataFrame with a unique identifier, scale information, statistics, and positions for each feature.
        drop : bool, default=False
            If ``True``, allow dropping of already existing feature impact and feature importance columns from
            ``df_feat`` before inserting.
        sample_positions : int, list of int, or None
            Position index/indices for the sample(s) in ``shap_values``.
            If ``None``, the impact for each sample will be returned.
        names: str or list of str, optional
            Unique name(s) used for the feature impact columns. When provided, they should align with
            ``sample_positions`` as follows:

            - **Single sample**: ``name`` as string and ``sample_positions`` as integer.
            - **Multiple samples**: ``name`` as list of string and ``sample_positions`` as corresponding list of integers.
            - **Group**: ``name`` as string and ``sample_positions`` as list of integers, each indicating a group sample.

            If ``sample_positions`` is ``None`` (all samples are considered), ``name`` must be list with names for each sample.
        normalize : bool, default=True
            If ``True``, normalize the feature impact to percentage.
        group_average : bool, default=False
            If ``True``, compute the average of samples given by ``sample_positions``.
        shap_feat_importance : bool, default=False
            If ``True``, include feature importance (i.e., absolute average SHAP values) instead of impact to ``df_feat``.

        Returns
        -------
        df_feat: pd.DataFrame, shape (n_features, n_feature_info+n)
            Feature DataFrame including feature impact. If the feature impact is computed for multiple samples,
            n=number of samples; n=1, otherwise.

        Notes
        -----
        **Feature impact (sample-level)**:
        The feature impact quantifies the positive or negative contribution of a feature to increase or decrease the model
        output for a specific sample (typically, prediction score). For each sample, the impact of an individual feature
        is represented by its corresponding SHAP value. These values are normalized such that the sum of their absolute
        values equals 100%.

        **Feature impact (group-level)**:
        The feature impact calculated for individual samples can be averaged to determine the feature impact for a group.
        This reflects how features influence the model's output on average within that group.

        **Feature importance (SHAP value-based)**:
        The average of the feature impact across all samples is termed as shap value-based 'feature importance'.
        This quantifies the overall contribution of each feature across the entire dataset.

        Warnings
        --------
        * If ``group_average=True``, warning when the standard deviation of a feature's impact significantly exceeds
          its mean impact, this may indicate an unreliable grouping.

        Examples
        --------
        .. include:: examples/se_add_feat_impact.rst
        """
        # Check input
        check_shap_values(shap_values=self.shap_values)
        n_samples, n_features = self.shap_values.shape
        df_feat = ut.check_df_feat(df_feat=df_feat)
        ut.check_bool(name="drop", val=drop)
        check_names(names=names)
        ut.check_bool(name="group_average", val=group_average)
        sample_positions = check_sample_positions(sample_positions=sample_positions, n_samples=n_samples)
        sample_positions, names = check_match_sample_positions_names(sample_positions=sample_positions, names=names,
                                                                     group_average=group_average)
        ut.check_bool(name="normalize", val=normalize)
        ut.check_bool(name="shap_feat_importance", val=shap_feat_importance)
        check_match_sample_positions_group_average(sample_positions=sample_positions, group_average=group_average)
        check_match_df_feat_shap_values(df_feat=df_feat, drop=drop, shap_values=self.shap_values,
                                        shap_feat_importance=shap_feat_importance)
        # Compute feature importance
        if shap_feat_importance:
            feat_importance = comp_shap_feature_importance(shap_values=self.shap_values,
                                                           normalize=normalize)
            df_feat = insert_shap_feature_importance(df_feat=df_feat,
                                                     feat_importance=feat_importance,
                                                     drop=drop)
        # Compute feature impact
        else:
            feat_impact = comp_shap_feature_impact(self.shap_values,
                                                   normalize=normalize,
                                                   sample_positions=sample_positions,
                                                   verbose=self._verbose,
                                                   group_average=group_average)
            df_feat = insert_shap_feature_impact(df_feat=df_feat,
                                                 feat_impact=feat_impact,
                                                 names=names,
                                                 drop=drop,
                                                 group_average=group_average)

        return df_feat

    @staticmethod
    def add_sample_mean_dif(X: ut.ArrayLike2D,
                            labels: ut.ArrayLike1D = None,
                            label_ref: int = 0,
                            df_feat: pd.DataFrame = None,
                            drop: bool = False,
                            sample_positions: Union[int, List[int], None] = None,
                            names: Optional[Union[str, List[str]]] = None,
                            group_average: bool = False,
                            ) -> pd.DataFrame:
        """
        Compute the feature value difference between selected samples and a reference group average.

        Three different scenarios for computing the difference with the reference group average (MEAN_REF) are possible:

            a) **Single sample**: Computes the difference for a selected sample and MEAN_REF.
            b) **Multiple samples**: Computes differences for multiple selected samples (all by default) individually against MEAN_REF.
            c) **Group of samples**: Computes the difference between the average of a selected group of samples and MEAN_REF.

        The calculated differences are added to ``df_feat`` as new columns named ``mean_dif_'name(s)'``,
        corresponding to each sample or group.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix. `Rows` typically correspond to proteins and `columns` to features.
        labels : array-like, shape (n_samples)
            Class labels for samples in ``X`` (typically, 1=positive, 0=negative).
        label_ref : int, default=0,
            Class label of reference group in ``labels``.
        df_feat : pd.DataFrame, shape (n_features, n_feature_info)
            Feature DataFrame with a unique identifier, scale information, statistics, and positions for each feature.
        drop : bool, default=False
            If ``True``, allow dropping of already existing sample specific mean difference columns from
            ``df_feat`` before inserting.
        sample_positions : int, list of int, or None
            Position index/indices for the sample(s) in ``shap_values``.
            If ``None``, the impact for each sample will be returned.
        names: str or list of str, optional
            Unique name(s) used for the feature value differences columns. When provided, they should align with
            ``sample_positions`` as follows:

            - **Single sample**: ``name`` as string and ``sample_positions`` as integer.
            - **Multiple samples**: ``name`` as list of string and ``sample_positions`` as corresponding list of integers.
            - **Group**: ``name`` as string and ``sample_positions`` as list of integers, each indicating a group sample.

            If ``sample_positions`` is ``None`` (all samples are considered), ``name`` must be list with names for each sample.
        group_average : bool, default=False
            If ``True``, compute the average of samples given by ``sample_positions``.

        Returns
        -------
        df_feat: pd.DataFrame, shape (n_features, n_feature_info+n)
            Feature DataFrame including feature value difference. If the feature value difference is computed for multiple
            samples, n=number of samples; n=1, otherwise.

        Examples
        --------
        .. include:: examples/se_add_sample_mean_dif.rst
        """
        # Check input
        X = ut.check_X(X=X)
        n_samples, n_feat = X.shape
        ut.check_X_unique_samples(X=X, min_n_unique_samples=2)
        ut.check_number_val(name="label_ref", val=label_ref, just_int=True)
        labels = ut.check_labels(labels=labels, vals_requiered=[label_ref],
                                 len_requiered=n_samples, allow_other_vals=True,
                                 accept_float=True) # Accept fuzzy labeling by default
        df_feat = ut.check_df_feat(df_feat=df_feat)
        ut.check_bool(name="drop", val=drop)
        check_names(names=names)
        ut.check_bool(name="group_average", val=group_average)
        sample_positions = check_sample_positions(sample_positions=sample_positions, n_samples=n_samples)
        sample_positions, names = check_match_sample_positions_names(sample_positions=sample_positions, names=names,
                                                                     group_average=group_average)
        # Compute differences
        df_feat = add_sample_mean_dif_(X, labels=labels, label_ref=label_ref,
                                       df_feat=df_feat, drop=drop,
                                       sample_positions=sample_positions, names=names,
                                       group_average=group_average)
        return df_feat



"""This is a script for utility functions for utility functions."""

# Helper functions
# This function is only used in utility check function
def add_str(str_error=None, str_add=None):
    """Add additional error message 'str_add' to default error message ('add_str')"""
    if str_add:
        str_error += "\n  " + str_add
    return str_error



"""
This is a script for utility feature statistics functions for CPP and SequenceFeature objects and backend.
"""
import numpy as np
from scipy import stats
from statsmodels.stats.multitest import multipletests

import aaanalysis.utils as ut


# I Helper Functions
# Summary and test statistics for feature matrix based on classification by labels
def _mean_dif(X=None, labels=None, label_test=1, label_ref=0):
    """ Get mean difference for values in X (feature matrix) based on y (labels)"""
    mask_test = [x == label_test for x in labels]
    mask_ref = [x == label_ref for x in labels]
    mean_difs = np.mean(X[mask_test], axis=0) - np.mean(X[mask_ref], axis=0)
    return mean_difs


def _std(X=None, labels=None, group=1):
    """Get standard deviation (std) for datasets points with group label"""
    mask = [x == group for x in labels]
    group_std = np.std(X[mask], axis=0)
    return group_std


def _p_correction(p_vals=None, p_cor="fdr_bh"):
    """Correct p-values"""
    # Exclude nan from list of corrected p-values
    p_vals_without_na = [p for p in p_vals if str(p) != "nan"]
    p_corrected_without_na = list(multipletests(p_vals_without_na, method=p_cor)[1])
    # Include nan in list of corrected p-values
    p_corrected = []
    i = 0
    for p in p_vals:
        if str(p) != "nan":
            p_corrected.append(p_corrected_without_na[i])
            i += 1
        else:
            p_corrected.append(np.nan)
    return p_corrected


def _mean_stat(X=None, labels=None, parametric=False, p_cor=None, label_test=1, label_ref=0):
    """Statistical comparison of central tendency between two groups for each feature"""
    mask_test = [x == label_test for x in labels]
    mask_ref = [x == label_ref for x in labels]
    if parametric:
        p_vals = stats.ttest_ind(X[mask_test], X[mask_ref], nan_policy="omit")[1]
        p_str = "p_val_ttest_indep"
    else:
        t = lambda x1, x2: stats.mannwhitneyu(x1, x2, alternative="two-sided")[1]  # Test statistic
        c = lambda x1, x2: np.mean(x1) != np.mean(x2) or np.std(x1) != np.std(x2)  # Test condition
        p_vals = np.round([t(col[mask_test], col[mask_ref]) if c(col[mask_test], col[mask_ref]) else 1 for col in X.T], 10)
        p_str = "p_val_mann_whitney"
    if p_cor is not None:
        p_vals = _p_correction(p_vals=p_vals, p_cor=p_cor)
        p_str = "p_val_" + p_cor
    return p_vals, p_str


# II Main Functions
def add_stat_(df=None, X=None, labels=None, parametric=False, label_test=1, label_ref=0):
    """Add summary statistics of feature matrix (X) for given labels (y) to df"""
    df = df.copy()
    columns_input = list(df)
    args_labels = dict(labels=labels, label_test=label_test, label_ref=label_ref)
    df[ut.COL_ABS_AUC] = abs(ut.auc_adjusted_(X=X, labels=labels, label_test=label_test))
    df[ut.COL_MEAN_DIF] = _mean_dif(X=X, **args_labels)
    if ut.COL_ABS_MEAN_DIF not in list(df):
        df[ut.COL_ABS_MEAN_DIF] = abs(_mean_dif(X=X, **args_labels))
    df[ut.COL_STD_TEST] = _std(X=X, labels=labels, group=label_test)
    df[ut.COL_STD_REF] = _std(X=X, labels=labels, group=label_ref)
    p_val, p_str = _mean_stat(X=X, parametric=parametric,**args_labels)
    df[p_str] = p_val
    p_val_fdr, p_str_fdr = _mean_stat(X=X, parametric=parametric, p_cor="fdr_bh", **args_labels)
    df[p_str_fdr] = p_val_fdr
    cols_stat = [ut.COL_ABS_AUC,
                 ut.COL_ABS_MEAN_DIF, ut.COL_MEAN_DIF,
                 ut.COL_STD_TEST, ut.COL_STD_REF,
                 p_str, p_str_fdr]
    cols_in = [x for x in columns_input if x not in cols_stat and x != ut.COL_FEATURE]
    columns = [ut.COL_FEATURE] + cols_in + cols_stat
    df = df[columns]
    cols_round = [x for x in cols_stat if x not in [p_str, p_str_fdr]]
    df[cols_round] = df[cols_round].round(6)
    return df


"""
This is a script for the backend of various AAclust methods.
"""
import pandas as pd
import numpy as np

import aaanalysis.utils as ut
from ._utils_aaclust import _compute_centers, _compute_medoids


# I Helper function
# Name clusters
def _get_cluster_names(list_names=None, name_medoid=None,
                       name_unclassified="Unclassified",
                       shorten_names=True):
    """
    Get list of cluster names sorted based on following criteria (descending order):
        a) Frequency of term (most frequent term is preferred)
        b) Term is the name or a sub-name of the given medoid
        c) Length of term (shorter terms are preferred)
    If cluster consists of only one term, the name will be 'unclassified ('category name')'
    """
    def remove_2nd_info(name_):
        """Remove information given behind comma"""
        if "," in name_:
            name_ = name_.split(",")[0]
            if "(" in name_:
                name_ += ")"  # Close parenthesis if interpreted by deletion
        return name_
    # Filter categories (Remove unclassified scales and secondary infos)
    list_names = [remove_2nd_info(x) for x in list_names if ut.STR_UNCLASSIFIED not in x]
    # Create list of shorter names not containing information given in parentheses
    list_short_names = [x.split(" (")[0] for x in list_names if " (" in x]
    if len(list_names) > 1:
        if shorten_names:
            list_names.extend(list_short_names)
        # Obtain information to check criteria for sorting scale names
        df_counts = pd.Series(list_names).value_counts().to_frame().reset_index()   # Compute frequencies of names
        df_counts.columns = ["name", "count"]
        df_counts["medoid"] = [True if x in name_medoid else False for x in df_counts["name"]]  # Name in medoid
        df_counts["length"] = [len(x) for x in df_counts["name"]]      # Length of name
        # Sort names based on given criteria
        df_counts = df_counts.sort_values(by=["count", "medoid", "length"], ascending=[False, False, True])
        names_cluster = df_counts["name"].tolist()
    else:
        names_cluster = [name_unclassified]
    return names_cluster


# Compute correlation
def _sort_X_labels_names(X, labels=None, names=None):
    """Sort labels"""
    sorted_order = np.argsort(labels)
    labels = np.array([labels[i] for i in sorted_order])
    X = X[sorted_order]
    if names:
        names = [names[i] for i in sorted_order]
    return X, labels, names

def _get_df_corr(X=None, X_ref=None):
    """Get df with correlations"""
    # Temporary labels to avoid any confusion with potential duplicates
    X_labels = range(len(X))
    X_ref_labels = range(len(X), len(X) + len(X_ref))
    combined = np.vstack((X, X_ref))
    df_corr_full = pd.DataFrame(combined.T).corr()
    # Select only the rows corresponding to X and columns corresponding to X_ref
    df_corr = df_corr_full.loc[X_labels, X_ref_labels]
    return df_corr


# II Main Functions
def compute_centers(X, labels=None):
    """Obtain cluster centers and their labels"""
    # Function in utilis for not breaking dependency rules:
    # Backend functions should only depend on backend utility functions
    return _compute_centers(X, labels=labels)


def compute_medoids(X, labels=None, metric="correlation"):
    """Obtain cluster medoids and their labels"""
    if metric is None:
        metric = "correlation"
    # Function in utilis for not breaking dependency rules:
    # Backend functions should only depend on backend utility functions
    return _compute_medoids(X, labels=labels, metric=metric)


def name_clusters(X, labels=None, names=None, shorten_names=True):
    """Create ordered list of cluster names"""
    medoids, medoid_labels, medoid_ind = compute_medoids(X, labels=labels)
    dict_medoids = dict(zip(medoid_labels, medoid_ind))
    # Get cluster labels sorted in descending order of frequency
    labels_sorted = pd.Series(labels).value_counts().index
    # Assign names to cluster
    dict_cluster_names = {}
    for clust in labels_sorted:
        name_medoid = names[dict_medoids[clust]]
        list_names = [names[i] for i in range(0, len(names)) if labels[i] == clust]
        names_cluster = _get_cluster_names(list_names=list_names, name_medoid=name_medoid,
                                           name_unclassified=ut.STR_UNCLASSIFIED,
                                           shorten_names=shorten_names)
        assigned = False
        for name in names_cluster:
            if name not in dict_cluster_names.values() or name == ut.STR_UNCLASSIFIED:
                dict_cluster_names[clust] = name
                assigned = True
                break
        if not assigned:
            dict_cluster_names[clust] = ut.STR_UNCLASSIFIED
    cluster_names = [dict_cluster_names[label] for label in labels]
    return cluster_names


def compute_correlation(X, X_ref=None, labels=None, labels_ref=None, names=None, names_ref=None):
    """Computes Pearson correlation of given data with reference data."""
    # Sort based on labels
    X, labels_sorted, names_sorted = _sort_X_labels_names(X, labels=labels, names=names)
    if X_ref is not None:
        X_ref, labels_ref, names_ref = _sort_X_labels_names(X_ref, labels=labels_ref, names=names_ref)
    # Compute correlations
    if X_ref is None:
        df_corr = pd.DataFrame(X.T).corr()
    else:
        df_corr = _get_df_corr(X=X, X_ref=X_ref)
    # Replace indexes and columns with names or labels
    df_corr.index = names_sorted if names else labels_sorted
    if X_ref is None:
        df_corr.columns = names_sorted if names else labels_sorted
    else:
        df_corr.columns = names_ref if names_ref else labels_ref
    return df_corr, labels_sorted



"""
This is a script for the backend of the CPP.run() method.

This is the key algorithm of CPP and for AAanalysis.
"""
import os
import numpy as np
import pandas as pd
import multiprocessing as mp
from itertools import repeat
import warnings

import aaanalysis.utils as ut
from .utils_feature import get_vf_scale, get_feature_matrix_, post_check_vf_scale
from ._utils_feature_stat import add_stat_
from ._split import SplitRange


# I Helper functions
def _get_splits(df_parts=None, split_type=None, split_type_args=None):
    """Split all given parts from df_parts. The split is given by its split_type
     (Segment, Pattern, or PeriodicPattern) and its arguments (split_kwargs)."""
    list_parts = list(df_parts)
    spr = SplitRange()
    f = getattr(spr, split_type.lower())
    if split_type_args is not None:
        f_splitr = lambda x: f(seq=x, **split_type_args)
        labels_s = getattr(spr, "labels_" + split_type.lower())(**split_type_args)
    else:
        f_splitr = lambda x: f(seq=x)   # Using default arguments
        labels_s = getattr(spr, "labels_" + split_type.lower())()
    # Combine part and split to get sequence splits
    part_splits = np.empty([len(df_parts), len(labels_s) * len(list_parts)], dtype=object)
    labels_ps = []  # Part Split labels
    for i, p in enumerate(list_parts):
        labels_ps.extend(["{}-{}".format(p.upper(), s) for s in labels_s])
        for j, seq in enumerate(df_parts[p]):
            part_splits[j, i*len(labels_s):(i+1)*len(labels_s)] = f_splitr(seq)
    return part_splits, labels_ps


# Pre-filtering and filtering
def _splitting(df_parts=None, split_kws=None):
    """Combine Part + Split.
    Returns:
        part_split: sequence fragments generated by applying splits on given sequence parts
        list_labels_ps: list with labels for Part-Split combination
    """
    list_parts_splits = []
    list_labels_ps = []
    for split_type in split_kws:
        split_type_args = split_kws[split_type]
        part_splits, labels_ps = _get_splits(df_parts=df_parts,
                                             split_type=split_type,
                                             split_type_args=split_type_args)
        list_parts_splits.append(part_splits)
        list_labels_ps.extend(labels_ps)
    part_split = np.concatenate(list_parts_splits, axis=1)
    return part_split, list_labels_ps


def _pre_filtering_info(list_scales, dict_all_scales, labels_ps, part_split, accept_gaps, mask_0, mask_1, verbose):
    """Compute abs(mean_dif) and std(test) to rank features, where mean_dif is the difference
    between the means of the test and the reference protein groups for a feature"""
    feat_names = np.empty((len(list_scales) * len(labels_ps)), dtype=object)
    abs_mean_dif = np.empty((len(list_scales) * len(labels_ps)))
    std_test = np.empty((len(list_scales) * len(labels_ps)))
    for i, scale in enumerate(list_scales):
        if verbose:
            ut.print_progress(i=i, n=len(list_scales))
        dict_scale = dict_all_scales[scale]
        start, end = i*len(labels_ps), (i+1)*len(labels_ps)
        # Feature names
        feat_names[start:end] = ["{}-{}".format(ps, scale) for ps in labels_ps]
        # Feature matrix
        vf_scale = get_vf_scale(dict_scale=dict_scale, accept_gaps=accept_gaps)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)    # Filter numpy warning: "Mean of emtpy slice"
            X = np.round(vf_scale(part_split), 5)
        if accept_gaps:
            post_check_vf_scale(feature_values=X)
        # Ranking infos
        abs_mean_dif[start:end] = abs(np.mean(X[mask_1], axis=0) - np.mean(X[mask_0], axis=0))
        std_test[start:end] = np.std(X[mask_1], axis=0)
    return abs_mean_dif, std_test, feat_names


def _filtering_info(df=None, df_scales=None, check_cat=True):
    """Get datasets structures for filtering, two dictionaries with feature to scale category resp.
    feature positions and one datasets frame with paired pearson correlations of all scales"""
    if check_cat:
        dict_c = dict(zip(df[ut.COL_FEATURE], df["category"]))
    else:
        dict_c = dict()
    dict_p = dict(zip(df[ut.COL_FEATURE], [set(x) for x in df["positions"]]))
    df_cor = df_scales.corr()
    return dict_c, dict_p, df_cor


# II Main functions
# Filtering methods
def pre_filtering_info(df_parts=None, split_kws=None, df_scales=None, labels=None, label_test=1, label_ref=0,
                       accept_gaps=False, verbose=True, n_jobs=None):
    """Get n best features in descending order based on the abs(mean(group1) - mean(group0),
    where group 1 is the target group"""
    # Input (df_parts, split_kws, df_scales, y) checked in main method (CPP.run())
    mask_ref = [x == label_ref for x in labels]
    mask_test = [x == label_test for x in labels]
    part_split, labels_ps = _splitting(split_kws=split_kws, df_parts=df_parts)
    list_scales = list(df_scales)
    dict_all_scales = {col: dict(zip(df_scales.index.to_list(), df_scales[col])) for col in list_scales}
    # Feature filtering
    if n_jobs == 1:
        # Run in a single process
        args = [list_scales, dict_all_scales, labels_ps, part_split, accept_gaps, mask_ref, mask_test, verbose]
        abs_mean_dif, std_test, feat_names = _pre_filtering_info(*args)
    else:
        # Run in multiple processes
        n_jobs = min([os.cpu_count(), len(list_scales)])
        scale_chunks = np.array_split(list_scales, n_jobs)
        args = zip(scale_chunks, repeat(dict_all_scales), repeat(labels_ps), repeat(part_split), repeat(accept_gaps),
                   repeat(mask_ref), repeat(mask_test), repeat(verbose))
        with mp.get_context("spawn").Pool(processes=n_jobs) as pool:
            result = pool.starmap(_pre_filtering_info, args)
        abs_mean_dif = np.concatenate([x[0] for x in result])
        std_test = np.concatenate([x[1] for x in result])
        feat_names = np.concatenate([x[2] for x in result])
    return abs_mean_dif, std_test, feat_names

def pre_filtering(features=None, abs_mean_dif=None, std_test=None, max_std_test=0.2, n=10000):
    """CPP pre-filtering based on thresholds."""
    df = pd.DataFrame(zip(features, abs_mean_dif, std_test),
                      columns=[ut.COL_FEATURE, ut.COL_ABS_MEAN_DIF, ut.COL_STD_TEST])
    df = df[df[ut.COL_STD_TEST] <= max_std_test]
    df = df.sort_values(by=ut.COL_ABS_MEAN_DIF, ascending=False).head(n)
    return df


def filtering(df=None, df_scales=None, max_overlap=0.5, max_cor=0.5, n_filter=100, check_cat=True):
    """CPP filtering algorithm based on redundancy reduction in descending order of absolute AUC."""
    dict_c, dict_p, df_cor = _filtering_info(df=df, df_scales=df_scales, check_cat=check_cat)
    df = df.sort_values(by=[ut.COL_ABS_AUC, ut.COL_ABS_MEAN_DIF], ascending=False).copy().reset_index(drop=True)
    list_feat = list(df[ut.COL_FEATURE])
    list_top_feat = [list_feat.pop(0)]  # List with best feature
    for feat in list_feat:
        add_flag = True
        # Stop condition for limit
        if len(list_top_feat) == n_filter:
            break
        # Compare features with all top features (added if low overlap & weak correlation or different category)
        for top_feat in list_top_feat:
            # If check_cat is False, the categories are not compared and only the position and correlation are checked
            if not check_cat or dict_c[feat] == dict_c[top_feat]:
                # Remove if feat positions high overlap or subset
                pos, top_pos = dict_p[feat], dict_p[top_feat]
                overlap = len(top_pos.intersection(pos))/len(top_pos.union(pos))
                if overlap >= max_overlap or pos.issubset(top_pos):
                    # Remove if high pearson correlation
                    scale, top_scale = feat.split("-")[2], top_feat.split("-")[2]
                    cor = df_cor[top_scale][scale]
                    if cor > max_cor:
                        add_flag = False
        if add_flag:
            list_top_feat.append(feat)
    df_top_feat = df[df[ut.COL_FEATURE].isin(list_top_feat)]
    return df_top_feat


# Adder methods for CPP analysis (used in run method)
def add_stat(df_feat=None, df_parts=None, df_scales=None, labels=None, parametric=False, accept_gaps=False,
             label_test=1, label_ref=0, n_jobs=None):
        """
        Add summary statistics for each feature to DataFrame.

        Notes
        -----
        P-values are calculated Mann-Whitney U test (non-parametric) or T-test (parametric) as implemented in SciPy.
        For multiple hypothesis correction, the Benjamini-Hochberg FDR correction is applied on all given features
        as implemented in SciPy.
        """
        # Add feature statistics
        features = list(df_feat[ut.COL_FEATURE])
        X = get_feature_matrix_(features=features,
                                df_parts=df_parts,
                                df_scales=df_scales,
                                accept_gaps=accept_gaps,
                                n_jobs=n_jobs)
        df_feat = add_stat_(df=df_feat, X=X, labels=labels, parametric=parametric,
                            label_test=label_test, label_ref=label_ref)
        return df_feat



"""
This is a script for getting current font size of figures.
"""
import seaborn as sns

# Main function
def plot_gcfs(option: str = 'font.size') -> int:
    """
    Get the current font size (or axes linewidth).

    This font size can be set by :func:`plot_settings` function.

    Parameters
    ----------
    option : str, default='font.size'
        Figure setting to get default value from. Either 'font.size' or 'axes.linewidth'

    Returns
    -------
    option_value : int
        Numerical value for selected option.

    See Also
    --------
    * `Plotting Prelude <plotting_prelude.html>`_.

    Examples
    --------
    .. include:: examples/plot_gcfs.rst
    """
    # Check input
    allowed_options = ["font.size", "axes.linewidth"]
    if option not in allowed_options:
        raise ValueError(f"'option' should be one of following: {allowed_options}")
    # Get the current plotting context
    current_context = sns.plotting_context()
    option_value = current_context[option]
    return option_value



"""
This is a script for the backend of the CPP.eval() method.
"""
import pandas as pd
import numpy as np
from sklearn.cluster import KMeans
import warnings

import aaanalysis.utils as ut
from .utils_feature import get_feature_matrix_


# Helper function
def _min_cor_center(X):
    """Get minimum for correlation of all columns with cluster center, defined as the mean values
    for each amino acid over all scales."""
    # Get cluster center
    f = lambda x: x.mean(axis=0)[np.newaxis, :]
    # Create new matrix including cluster center
    center_X = np.concatenate([f(X), X], axis=0)
    # Get minimum correlation with mean values
    min_cor = np.corrcoef(center_X)[0, ].min()
    return min_cor


def get_min_cor(X, labels=None):
    """Compute minimum pair-wise correlation for each cluster label and return minimum of obtained cluster minimums."""
    # Minimum correlations for each cluster (with center or all scales)
    list_masks = [[i == label for i in labels] for label in set(labels)]
    list_min_cor = [_min_cor_center(X[mask]) for mask in list_masks]
    # Minimum for all clusters
    min_cor = min(list_min_cor)
    return min_cor


def get_best_n_clusters(X=None, min_th=0.3, random_state=None):
    """Obtain the best number of clusters based on internal cluster minimum correlation."""
    n_features, _ = X.shape
    max_n = min([100, n_features-1])
    best_n_clusters = max_n
    for n_clusters in range(2, max_n):
        kmeans = KMeans(n_clusters=n_clusters, n_init="auto", random_state=random_state)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            labels = kmeans.fit_predict(X)
            min_cor = get_min_cor(X, labels=labels)
        # Check if the minimum correlation in all clusters is at least 0.3
        if min_cor >= min_th:
            best_n_clusters = n_clusters
            break
    return best_n_clusters


def get_features_per_cluster_stats(X=None, n_clusters=2, random_state=None):
    """Calculate the average number of features per cluster and its standard deviation."""
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        kmeans = KMeans(n_clusters=n_clusters, n_init="auto", random_state=random_state)
        labels = kmeans.fit_predict(X)
    # Count features in each cluster
    feat_per_cluster = [np.sum(labels == i) for i in range(n_clusters)]
    # Calculate average and standard deviation
    avg_feat_per_cluster = round(np.mean(feat_per_cluster), 2)
    std_feat_per_cluster = round(np.std(feat_per_cluster), 2)
    return avg_feat_per_cluster, std_feat_per_cluster


# II Main function
@ut.catch_runtime_warnings()
def evaluate_features(list_df_feat=None, names_feature_sets=None, list_cat=None,
                      list_df_parts=None, df_scales=None, accept_gaps=False,
                      n_jobs=1, min_th=0.0, random_state=None):
    """Evaluate sets of clustering labels"""
    list_evals = []
    for df_feat, df_parts in zip(list_df_feat, list_df_parts):
        # Filter feature dataframe
        df = df_feat[df_feat[ut.COL_CAT].isin(list_cat)].copy()
        if len(df) == 0:
            raise ValueError(f"'list_cat' ({list_cat}) results in empty 'df_feat' after filtering.")
        eval_dict = {}
        # Get number features per of category
        eval_dict[ut.COL_N_FEAT] = (len(df), [len(df[df[ut.COL_CAT] == cat]) for cat in list_cat])
        # Compute CPP statistics
        eval_dict[ut.COL_AVG_ABS_AUC] = df[ut.COL_ABS_AUC].mean()
        eval_dict[ut.COL_RANGE_ABS_AUC] = df[ut.COL_ABS_AUC].quantile([0, 0.25, 0.5, 0.75, 1]).round(3).tolist()
        eval_dict[ut.COL_AVG_MEAN_DIF] = (round(df[df[ut.COL_MEAN_DIF] > 0][ut.COL_MEAN_DIF].mean(), 3),
                                          round(df[df[ut.COL_MEAN_DIF] < 0][ut.COL_MEAN_DIF].mean(), 3))
        # Obtain optimal number of clusters
        features = list(df[ut.COL_FEATURE])
        X = get_feature_matrix_(features=features, df_parts=df_parts, df_scales=df_scales,
                                accept_gaps=accept_gaps, n_jobs=n_jobs)
        # Get n_clusters for features
        best_n_clust = get_best_n_clusters(X=X.T, min_th=min_th, random_state=random_state)
        eval_dict[ut.COL_N_CLUST] = best_n_clust
        avg_feat_per_cluster, std_feat_per_cluster = get_features_per_cluster_stats(X=X.T,
                                                                                    n_clusters=best_n_clust,
                                                                                    random_state=random_state)
        eval_dict[ut.COL_AVG_N_FEAT_PER_CLUST] = avg_feat_per_cluster
        eval_dict[ut.COL_STD_N_FEAT_PER_CLUST] = std_feat_per_cluster
        # Append this dictionary to the list_evals
        list_evals.append(eval_dict)
    # Create the DataFrame
    df_eval = pd.DataFrame(list_evals).round(3)
    df_eval = ut.add_names_to_df_eval(df_eval=df_eval, names=names_feature_sets)
    return df_eval




"""
This is a script for frontend of the setting plot legend.
The backend is in general utility module to provide function to remaining AAanalysis modules.
"""
from typing import Optional, List, Dict, Union, Tuple
from matplotlib import pyplot as plt
from aaanalysis import utils as ut

# I Helper functions


# II Main function
def plot_legend(ax: Optional[plt.Axes] = None,
                # Categories and colors
                dict_color: Dict[str, str] = None,
                list_cat: Optional[List[str]] = None,
                labels: Optional[List[str]] = None,
                # Position and Layout
                loc: Union[str, int] = "upper left",
                loc_out: bool = False,
                frameon: bool = False,
                y: Optional[Union[int, float]] = None,
                x: Optional[Union[int, float]] = None,
                n_cols: int = 3,
                labelspacing: Union[int, float] = 0.2,
                columnspacing: Union[int, float] = 1.0,
                handletextpad: Union[int, float] = 0.8,
                handlelength: Union[int, float] = 2.0,
                # Font and Style
                fontsize: Optional[Union[int, float]] = None,
                fontsize_title: Optional[Union[int, float]] = None,
                weight_font: str = "normal",
                weight_title: str = "normal",
                # Marker, Lines, and Area
                marker: Optional[Union[str, int, list]] = None,
                marker_size: Union[int, float, List[Union[int, float]]] = 10,
                lw: Union[int, float] = 0,
                linestyle: Optional[Union[str, list]] = None,
                edgecolor: Optional[str] = None,
                hatch: Optional[Union[str, List[str]]] = None,
                hatchcolor: str = "white",
                # Title
                title: Optional[str] = None,
                title_align_left: bool = True,
                **kwargs
                ) -> Union[plt.Axes, Tuple[List, List[str]]]:
    """
    Set an independently customizable plot legend.

    Legends can be flexibly adjusted based categories and colors provided in ``dict_color`` dictionary.
    This functions comprises the most convenient settings for ``func:`matplotlib.pyplot.legend``.

    Parameters
    ----------
    ax : plt.Axes, optional
        The axes to attach the legend to. If not provided, the current axes will be used.
    dict_color : dict, optional
        A dictionary mapping categories to colors.
    list_cat : list of str, optional
        List of categories to include in the legend (keys of ``dict_color``).
    labels : list of str, optional
        Legend labels corresponding to given categories.
    loc : int or str
        Location for the legend.
    loc_out : bool, default=False
        If ``True``, sets automatically ``x=0`` and ``y=-0.25`` if they are ``None``.
    frameon : bool, default=False
        If ``True``, a figure background patch (frame) will be drawn.
    y : int or float, optional
        The y-coordinate for the legend's anchor point.
    x : int or float, optional
        The x-coordinate for the legend's anchor point.
    n_cols : int, default=1
        Number of columns in the legend, at least 1.
    labelspacing : int or float, default=0.2
        Vertical spacing between legend items.
    columnspacing : int or float, default=1.0
        Horizontal spacing between legend columns.
    handletextpad : int or float, default=0.8
        Horizontal spacing between legend handle (marker) and label.
    handlelength : int or float, default=2.0
        Length of legend handle.
    fontsize : int or float, optional
        Font size of the legend text.
    fontsize_title : inf or float, optional
        Font size of the legend title.
    weight_font : str, default='normal'
        Weight of the font.
    weight_title : str, default='normal'
        Font weight for the legend title.
    marker : str, int, or list, optional
        Handle marker for legend items. Lines ('-') only visible if ``lw>0``.
    marker_size : int, float, or list, optional
        Marker size of legend items.
    lw : int or float, default=0
        Line width for legend items. If negative, corners are rounded.
    linestyle : str or list, optional
        Style of line. Only applied to lines (``marker='-'``).
    edgecolor : str, optional
        Edge color of legend items. Not applicable to lines.
    hatch : str or list, optional
        Filling pattern for default marker. Only applicable when ``marker=None``.
    hatchcolor : str, default='white'
        Hatch color of legend items. Only applicable when ``marker=None``.
    title : str, optional
        Legend title.
    title_align_left : bool, default=True
        Whether to align the title to the left.
    **kwargs
        Further key word arguments for :attr:`matplotlib.axes.Axes.legend`.

    Returns
    -------
    ax : plt.Axes
        The axes object on which legend is applied to.

    Notes
    -----
    Markers can be None (default), lines ('-') or one of the `matplotlib markers
    <https://matplotlib.org/stable/api/markers_api.html>`_.

    See Also
    --------
    * More examples in `Plotting Prelude <plotting_prelude.html>`_.
    * `Linestyles of markers <https://matplotlib.org/stable/gallery/lines_bars_and_markers/linestyles.html>`_.
    * `Hatches <https://matplotlib.org/stable/gallery/shapes_and_collections/hatch_style_reference.html>`_, which are filling patterns.
    * :class:`matplotlib.lines.Line2D` for available marker shapes and line properties.
    * :class:`matplotlib.axes.Axes`, which is the core object in matplotlib.
    * :func:`matplotlib.pyplot.gca` to get the current Axes instance.

    Examples
    --------
    .. include:: examples/plot_legend.rst
    """
    # Check input
    ut.check_ax(ax=ax, accept_none=True)
    if ax is None:
        ax = plt.gca()
    ut.check_dict(name="dict_color", val=dict_color, accept_none=False)
    ut.check_bool(name="title_align_left", val=title_align_left)
    ut.check_bool(name="loc_out", val=loc_out)
    ut.check_bool(name="frameon", val=frameon)
    ut.check_number_range(name="n_cols", val=n_cols, min_val=1, accept_none=True, just_int=True)
    ut.check_number_val(name="x", val=x, accept_none=True, just_int=False)
    ut.check_number_val(name="y", val=y, accept_none=True, just_int=False)
    ut.check_number_val(name="lw", val=lw, accept_none=True, just_int=False)
    args_non_neg = {"labelspacing": labelspacing, "columnspacing": columnspacing,
                    "handletextpad": handletextpad, "handlelength": handlelength,
                    "fontsize": fontsize, "fontsize_legend": fontsize_title}
    for key in args_non_neg:
        ut.check_number_range(name=key, val=args_non_neg[key], min_val=0, accept_none=True, just_int=False)
    # Create new legend
    ax = ut.plot_legend_(ax=ax, dict_color=dict_color, list_cat=list_cat, labels=labels,
                         loc=loc, loc_out=loc_out, y=y, x=x, n_cols=n_cols,
                         labelspacing=labelspacing, columnspacing=columnspacing,
                         handletextpad=handletextpad, handlelength=handlelength,
                         fontsize=fontsize, fontsize_title=fontsize_title,
                         weight_font=weight_font, weight_title=weight_title,
                         marker=marker, marker_size=marker_size, lw=lw, linestyle=linestyle, edgecolor=edgecolor,
                         hatch=hatch, hatchcolor=hatchcolor, title=title, title_align_left=title_align_left,
                         frameon=frameon, **kwargs)
    return ax



from .data_handling import load_dataset, load_scales, load_features, to_fasta
from .feature_engineering import AAclust, AAclustPlot, SequenceFeature, NumericalFeature, CPP, CPPPlot
from .pu_learning import dPULearn, dPULearnPlot
from .explainable_ai import TreeModel, ShapExplainer
from .pertubation import AAMut, AAMutPlot, SeqMut, SeqMutPlot
from .plotting import (plot_get_clist, plot_get_cmap, plot_get_cdict,
                       plot_settings, plot_legend, plot_gcfs,
                       display_df)
from .metrics import (comp_auc_adjusted, comp_bic_score, comp_kld)
from .config import options

__all__ = [
    "load_dataset",
    "load_scales",
    "load_features",
    "to_fasta",
    "AAclust",
    "AAclustPlot",
    "SequenceFeature",
    "NumericalFeature",
    "CPP",
    "CPPPlot",
    "dPULearn",
    "dPULearnPlot",
    "TreeModel", "ShapExplainer",
    "AAMut",
    "AAMutPlot",
    "SeqMut",
    "SeqMutPlot",
    "plot_get_clist",
    "plot_get_cmap",
    "plot_get_cdict",
    "plot_settings",
    "plot_legend",
    "plot_gcfs",
    "display_df",
    "comp_auc_adjusted",
    "comp_bic_score",
    "comp_kld",
    "options"
]





"""
Script for (backend class) Split objects used to fragment sequence parts into distinct segments or patterns.
"""
import numpy as np
import itertools

import aaanalysis.utils as ut


# Pattern helper functions
def _get_pattern_pos(steps=None, repeat=2, len_max=12):
    """Get all possible positions from steps with number of repeats and maximum length using itertools"""
    list_steps = itertools.product(steps, repeat=repeat)    # Cartesian product of all step combinations
    list_pos = [np.cumsum(s) for s in list_steps]       # Positions from steps
    # Get all possible pattern positions
    list_pattern_pos = []
    for p in list_pos:
        max_p = max(p)
        min_p = min(p)
        if max_p <= len_max:
            for i in range(max_p-len_max, min_p):
                pattern_pos = list(p - i)
                if pattern_pos not in list_pattern_pos:
                    list_pattern_pos.append(pattern_pos)
    return list_pattern_pos


def get_list_pattern_pos(steps=None, n_min=2, n_max=4, len_max=15):
    """Get list of pattern positions using get_pattern_pos"""
    list_pattern_pos = []
    for n in range(n_min, n_max+1):
        list_pattern_pos.extend(_get_pattern_pos(steps=steps, repeat=n, len_max=len_max))
    list_pattern_pos = sorted(list_pattern_pos)
    return list_pattern_pos


# II Main Functions
class Split:
    """Class for splitting parts into Segments, Patterns, and PeriodicPatterns.
    Counting of amino acid positions for splits start at 1 and N-terminal of the
    protein sequence"""
    def __init__(self, type_str=True):
        self.type_str = type_str

    @staticmethod
    def segment(seq=None, i_th=1, n_split=2):
        """Get i-th segment of sequence that is split into n contiguous segments.
        For starting at the C-terminus, the sequence will be reversed.

        Parameters
        ----------
        seq: sequence (e.g., amino acids sequence for protein)
        i_th: integer indicating the selected segment
        n_split: integer indicating the number of segments to split the given sequence in

        Returns
        -------
        seq_segment: Segment split for given sequence

        Notes
        -----
        Segments are denoted as 'Segment(i-th,n_split)'
        """
        len_segment = len(seq) / n_split
        start = int(len_segment * (i_th - 1))   # Start at 0 for i_th = 1
        end = int(len_segment * i_th)
        # DEV: If 'IndexError: list index out of range',
        #  check in interface that sequence size matches with features
        #  via 'check_match_features_seq_parts'
        seq_segment = seq[start:end]
        return seq_segment

    def pattern(self, seq=None, terminus="N", list_pos=None):
        """Get sequence pattern consisting of n positions given in list_pos.By default, a pattern starts at
        the N-terminus. For starting at the C-terminus, the sequence will be reversed.

        Parameters
        ----------
        seq: sequence (e.g., amino acids sequence for protein)
        terminus: 'N' or 'C' indicating the start side of split
        list_pos: list with integers indicating the positions the Pattern split should consist of

        Returns
        -------
        seq_pattern: Pattern split for given sequence

        Notes
        -----
        Patterns are denoted as 'Pattern(N/C,p1,p2,...,pn)',
            where N or C specifies whether the pattern starts at the N- or C-terminus of the sequence
            and pn denotes the n-th position from list_pos.
        """
        list_pos = [int(x) for x in list_pos if x is not None]
        if terminus == "C":
            seq = seq[::-1]
        if self.type_str:
            seq_pattern = "".join([seq[i-1] for i in list_pos])
        else:
            # DEV: If 'IndexError: list index out of range',
            #  check in interface that sequence size matches with features
            #  via 'check_match_features_seq_parts'
            seq_pattern = [seq[i-1] for i in list_pos]
        return seq_pattern

    def periodicpattern(self, seq=None, terminus="N", step1=None, step2=None, start=1):
        """Get a periodic sequence pattern consisting of elements with an alternating step size given by step1, step2
        and starting at the given start position. For starting at the C-terminus, the sequence will be reversed.

        Parameters
        ----------
        seq: sequence (e.g., amino acids sequence for protein)
        terminus: 'N' or 'C' indicating the start side of split
        step1: integer indicating odd step sizes
        step2: integer indicating even step sizes
        start: integer indicating start position (starting from 1)

        Returns
        -------
        seq_periodicpattern: PeriodicPattern split for given sequence

        Notes
        -----
        PeriodicPatterns are denoted as 'periodicpattern(N/C,i+step1/step2,start)',
            where N or C specifies whether the pattern starts at the N- or C-terminus of the sequence,
            i+step1/step2 defines the alternating g step sizes, and start gives the start position beginning at 0.

        Giving an example for proteins, for step1=3, step2=4, start=1, the periodic pattern consists of
            alternating 3rd and 4th amino acids within the whole sequence starting at the position one.
            By default, a pattern starts at the N-terminus. For starting at the C-terminus, the sequence must
            be reversed. For IMP substrates, the periodic pattern is representing a face of the
            helical transmembrane domain given for step1, step2 in {3, 4} and start >= step1.
        """
        if terminus == "C":
            seq = seq[::-1]
        pos = start
        list_pos = [pos]
        while pos <= len(seq):
            if len(list_pos) % 2 != 0:
                pos += step1
            else:
                pos += step2
            if pos <= len(seq):
                list_pos.append(pos)
        if self.type_str:
            seq_periodicpattern = "".join([seq[i-1] for i in list_pos])
        else:
            # DEV: If 'IndexError: list index out of range',
            #  check in interface that sequence size matches with features
            #  via 'check_match_features_seq_parts'
            seq_periodicpattern = [seq[i-1] for i in list_pos]
        return seq_periodicpattern


class SplitRange:
    """Class for creating range of splits for testing sets of multiple features in CPP"""

    # Segment methods
    @staticmethod
    def segment(seq=None, n_split_min=1, n_split_max=15):
        """Get range of all possible Segment splits for given sequences."""
        sp = Split()
        f = sp.segment  # Unbound function for higher performance
        seq_splits = []
        for n_split in range(n_split_min, n_split_max+1):
            for i_th in range(1, n_split+1):
                seq_segment = f(seq=seq, n_split=n_split, i_th=i_th)
                seq_splits.append(seq_segment)
        return seq_splits

    @staticmethod
    def labels_segment(n_split_min=1, n_split_max=15):
        """Get labels for range of Segment splits."""
        labels = []
        for n_split in range(n_split_min, n_split_max+1):
            for i_th in range(1, n_split+1):
                name = "{}({},{})".format(ut.STR_SEGMENT, i_th, n_split)
                labels.append(name)
        return labels

    # Pattern methods
    @staticmethod
    def pattern(seq=None, steps=None, n_min=2, n_max=4, len_max=15):
        """Get range of all possible Pattern splits for given sequence."""
        list_pattern_pos = get_list_pattern_pos(steps=steps, n_min=n_min, n_max=n_max, len_max=len_max)
        sp = Split()
        f = sp.pattern  # Unbound function for higher performance
        seq_splits = []
        for terminus in ['N', 'C']:
            for pattern_pos in list_pattern_pos:
                seq_pattern = f(seq=seq, terminus=terminus, list_pos=pattern_pos)
                seq_splits.append(seq_pattern)
        return seq_splits

    @staticmethod
    def labels_pattern(steps=None, n_min=2, n_max=4, len_max=15):
        """Get labels for range of Pattern splits."""
        list_pattern_pos = get_list_pattern_pos(steps=steps, n_min=n_min, n_max=n_max, len_max=len_max)
        labels = []
        for terminus in ['N', 'C']:
            for pattern_pos in list_pattern_pos:
                name = "{}({},{})".format(ut.STR_PATTERN, terminus, ",".join(str(x) for x in pattern_pos))
                labels.append(name)
        return labels

    # Periodic pattern methods
    @staticmethod
    def periodicpattern(seq=None, steps=None):
        """Get range of all possible PeriodicPattern splits for given sequence"""
        sp = Split()
        f = sp.periodicpattern  # Unbound function for higher performance
        seq_splits = []
        for terminus in ['N', 'C']:
            for step1, step2 in itertools.product(steps, repeat=2):
                for start in range(1, step1+1):
                    seq_periodicpattern = f(seq=seq, terminus=terminus, step1=step1, step2=step2, start=start)
                    seq_splits.append(seq_periodicpattern)
        return seq_splits

    @staticmethod
    def labels_periodicpattern(steps=None):
        """Get labels of all possible PeriodicPattern splits."""
        labels = []
        for terminus in ['N', 'C']:
            for step1, step2 in itertools.product(steps, repeat=2):
                for start in range(1, step1+1):
                    name = "{}({},i+{}/{},{})".format(ut.STR_PERIODIC_PATTERN, terminus, step1, step2, start)
                    labels.append(name)
        return labels




"""This is a script for the backend of the TreeModel.eval() method."""
import pandas as pd
import numpy as np
from sklearn.model_selection import cross_val_score

import aaanalysis.utils as ut


# II Main methods
@ut.catch_undefined_metric_warning()
def eval_feature_selections(X, labels=None, list_is_selected=None, names_feature_selections=None, n_cv=5,
                            list_metrics=None, list_model_classes=None, list_model_kwargs=None, verbose=True):
    """Evaluate the performance of different feature selections for multiple models,
    and compute the average score across all models and rounds for each scoring metric."""
    n_feature_sets = len(list_is_selected)
    n_metrics = len(list_metrics)
    list_evals = np.empty((n_feature_sets, n_metrics))
    if verbose:
        ut.print_start_progress(start_message=f"Tree Model starts evaluation of {n_feature_sets} feature sets.")
    for i, is_selected_rounds in enumerate(list_is_selected):
        n_rounds = len(is_selected_rounds)
        results_round = np.empty((n_rounds, n_metrics))
        for j, is_selected in enumerate(is_selected_rounds):
            X_selected = X[:, is_selected]
            for k, eval_score in enumerate(list_metrics):
                if verbose:
                    pct_progress = (j + k/n_metrics)/n_rounds
                    ut.print_progress(i=i+pct_progress, n=n_feature_sets)
                model_scores = [
                    cross_val_score(model_class(**model_kwargs), X_selected, y=labels, cv=n_cv, scoring=eval_score).mean()
                    for model_class, model_kwargs in zip(list_model_classes, list_model_kwargs)
                ]
                results_round[j, k] = np.mean(model_scores)
        # Average the scores across all rounds for each scoring metric
        list_evals[i, :] = results_round.mean(axis=0)
    if verbose:
        ut.print_end_progress(end_message=f"Tree Model finished evaluation and saves results.")
    # Create the DataFrame
    cols_eval = [metric if isinstance(metric, str) else metric.__name__ for metric in list_metrics]
    df_eval = pd.DataFrame(list_evals, columns=cols_eval).round(4)
    df_eval = ut.add_names_to_df_eval(df_eval=df_eval, names=names_feature_selections)
    return df_eval



"""
This is a script for adjusting terminal output.
"""
import numpy as np
import sys

STR_PROGRESS = "."

# I Helper Functions
# Plotting & print functions
def _print_red(input_str, **args):
    """Prints the given string in red text."""
    print(f"\033[91m{input_str}\033[0m", **args)

def _print_blue(input_str, **args):
    """Prints the given string in blue text."""
    print(f"\033[94m{input_str}\033[0m", **args)

def _print_green(input_str, **args):
    """Prints the given string in Matrix-style green text."""
    print(f"\033[92m{input_str}\033[0m", **args)
    #print(f"\033[32m{input_str}\033[0m", **args)

def print_out(input_str, **args):
    """Prints the given string in Matrix-style green text."""
    _print_blue(input_str, **args)

# Progress bar
def print_start_progress(start_message=None):
    """Print start progress"""
    # Start message
    if start_message is not None:
        print_out(start_message)
    # Start progress bar
    progress_bar = " " * 25
    print_out(f"\r   |{progress_bar}| 0.0%", end="")
    sys.stdout.flush()


def print_progress(i=0, n=0, add_new_line=False):
    """Print progress"""
    progress = min(np.round(i/n * 100, 2), 100)
    progress_bar = STR_PROGRESS * int(progress/4) + " " * (25-int(progress/4))
    str_end = "\n" if add_new_line else ""
    print_out(f"\r   |{progress_bar}| {progress:.1f}%", end=str_end)
    sys.stdout.flush()


def print_end_progress(end_message=None):
    """Print finished progress bar"""
    # End progress bar
    progress_bar = STR_PROGRESS * 25
    print_out(f"\r   |{progress_bar}| 100.0%")
    # End message
    if end_message is not None:
        print_out(end_message)
    sys.stdout.flush()




"""
This is a script for the backend of the dPULearn.fit() method.
"""
import pandas as pd
import numpy as np
from sklearn.metrics import pairwise_distances
import math
from sklearn.decomposition import PCA

import aaanalysis.utils as ut

# II Main Functions
def get_neg_via_distance(X=None, labels=None, metric="euclidean", n_unl_to_neg=None,
                         label_neg=0, label_pos=1):
    """Identify distant samples from positive mean as reliable negatives based on a specified distance metric."""
    col_dif = f'{metric}_dif'
    col_dif_abs = f"{metric}_abs_dif"
    mask_pos = labels == label_pos
    mask_unl = labels != label_pos
    # Compute the distances to the average value of the positive datapoints
    dif_to_pos_mean = pairwise_distances(X[mask_pos], X, metric=metric).mean(axis=0)
    abs_dif_to_pos_mean = np.abs(dif_to_pos_mean)
    # Create a DataFrame with the distances
    df_pu = pd.DataFrame({col_dif: dif_to_pos_mean, col_dif_abs: abs_dif_to_pos_mean})
    # Select negatives based on largest average distance to positives
    top_indices = df_pu[mask_unl].sort_values(by=col_dif_abs).tail(n_unl_to_neg).index
    new_labels = labels.copy()
    new_labels[top_indices] = label_neg
    # Adjust df distance
    df_pu = df_pu.round(4)
    df_pu.insert(0, ut.COL_SELECTION_VIA, [metric if l == 0 else None for l in new_labels])
    return new_labels, df_pu


def get_neg_via_pca(X=None, labels=None, n_components=0.8, n_unl_to_neg=None,
                    label_neg=0, label_pos=1, **pca_kwargs):
    """Identify distant samples from positive mean as reliable negatives in PCA-compressed feature spaces."""
    # Principal component analysis
    pca = PCA(n_components=n_components, **pca_kwargs)
    pca.fit(X.T)
    list_exp_var = pca.explained_variance_ratio_
    columns_pca = [f"PC{n+1} ({round(exp_var*100, 1)}%)" for n, exp_var in enumerate(list_exp_var)]
    # Determine number of negatives based on explained variance
    _list_n_neg = [math.ceil(n_unl_to_neg * x / sum(list_exp_var)) for x in list_exp_var]
    _list_n_cumsum = np.cumsum(np.array(_list_n_neg))
    list_n_neg = [n for n, cs in zip(_list_n_neg, _list_n_cumsum) if cs <= n_unl_to_neg]
    if sum(list_n_neg) != n_unl_to_neg:
        list_n_neg.append(n_unl_to_neg - sum(list_n_neg))
    columns_pca = columns_pca[:len(list_n_neg)]
    # Create df_pu based on PCA components
    df_pu = pd.DataFrame(pca.components_.T[:, :len(columns_pca)], columns=columns_pca)
    # Get mean of positive data for each component
    mask_pos = labels == label_pos
    mask_unl = labels != label_pos
    pc_means = df_pu[mask_pos].mean(axis=0)
    # Select negatives based on absolute difference to mean of positives for each component
    df_pu.insert(0, ut.COL_SELECTION_VIA, None)  # New column to store the PC information
    new_labels = labels.copy()
    for col_pc, mean_pc, n in zip(columns_pca, pc_means, list_n_neg):
        col_abs_dif = f"{col_pc}_abs_dif"
        # Calculate absolute difference to the mean for each sample in the component
        df_pu[col_abs_dif] = np.abs(df_pu[col_pc] - mean_pc)
        # Sort and take top n indices
        top_indices = df_pu[mask_unl].sort_values(by=col_abs_dif).tail(n).index
        # Update labels and masks
        new_labels[top_indices] = label_neg
        mask_unl[top_indices] = False
        # Record the PC by which the negatives are selected
        df_pu.loc[top_indices, ut.COL_SELECTION_VIA] = col_pc.split(' ')[0]
    # Adjust df
    cols = [x for x in list(df_pu) if x != ut.COL_SELECTION_VIA]
    df_pu[cols] = df_pu[cols].round(4)
    return new_labels, df_pu




"""
This is a script for the backend of the ShapExplainer.fit() method.
"""
import numpy as np
import shap

import aaanalysis.utils as ut


LIST_VERBOSE_SHAP_EXPLAINERS = [shap.KernelExplainer]


# I Helper Functions
def _get_background_data(X, explainer_class=None, n_background_data=None):
    """Sample background data for Kernel explainer"""
    X = X.copy()
    if explainer_class == shap.KernelExplainer and n_background_data is not None:
        # Summarize background data using k-means if specified
        background_data = shap.kmeans(X, n_background_data)
    else:
        background_data = X
    return background_data

def _get_shap_values(shap_output, class_index=1):
    """Retrieve SHAP values for the specified class index from SHAP output."""
    # For other types of SHAP outputs (like those from specific explainers)
    if isinstance(shap_output, list) and len(shap_output) > class_index:
        shap_values = shap_output[class_index]
    else:
        shap_values = shap_output
    return shap_values


def _compute_shap_values(X, labels, model_class=None, model_kwargs=None,
                         explainer_class=None, explainer_kwargs=None,
                         class_index=1, n_background_data=None):
    """Fit a model and compute SHAP values using the provided SHAP Explainer."""
    # Fit the model
    model = model_class(**model_kwargs).fit(X, labels)
    # Instantiate explainer class
    if explainer_class in [shap.KernelExplainer, shap.DeepExplainer, shap.GradientExplainer]:
        background_data = _get_background_data(X, explainer_class=explainer_class, n_background_data=n_background_data)
        model_input = model.predict_proba if hasattr(model, 'predict_proba') else model.predict
        explainer = explainer_class(model_input, background_data, **explainer_kwargs)
    elif explainer_class == shap.LinearExplainer:
        explainer = explainer_class(model, X, **explainer_kwargs)
    else:
        explainer = explainer_class(model, **explainer_kwargs)

    # Compute SHAP values
    if 'y' in explainer.shap_values.__code__.co_varnames:
        shap_output = explainer.shap_values(X, y=labels)
    else:
        shap_output = explainer.shap_values(X)

    # Process SHAP values
    shap_values = _get_shap_values(shap_output, class_index=class_index)
    # Handle expected value for classification models
    expected_value = explainer.expected_value
    if not np.isscalar(expected_value):
        expected_value = expected_value[class_index] if len(expected_value) > class_index else expected_value[-1]
    return shap_values, expected_value


def _aggregate_shap_values(X, labels=None, list_model_classes=None, list_model_kwargs=None,
                           explainer_class=None, explainer_kwargs=None,
                           class_index=1, n_background_data=None):
    """Aggregate SHAP values across multiple models."""

    n_samples, n_features = X.shape
    n_models = len(list_model_classes)
    shap_values_rounds = np.empty(shape=(n_samples, n_features, n_models))
    list_expected_value = []
    for i, model_class in enumerate(list_model_classes):
        model_kwargs = list_model_kwargs[i]
        args = dict(model_class=model_class, model_kwargs=model_kwargs,
                    explainer_class=explainer_class, explainer_kwargs=explainer_kwargs,
                    class_index=class_index, n_background_data=n_background_data)
        shap_values, expected_value = _compute_shap_values(X, labels=labels, **args)
        shap_values_rounds[:, :, i] = shap_values
        list_expected_value.append(expected_value)
    shap_values = np.mean(shap_values_rounds, axis=2)
    exp_val = np.mean(list_expected_value)
    return shap_values, exp_val


# II Main Functions
def monte_carlo_shap_estimation(X, labels=None, list_model_classes=None, list_model_kwargs=None,
                                explainer_class=None, explainer_kwargs=None, n_rounds=5,
                                is_selected=None, fuzzy_labeling=False, verbose=False,
                                label_target_class=1, n_background_data=None):
    """Compute Monte Carlo estimates of SHAP values for multiple models and feature selections."""
    # Get class index
    label_classes = sorted(list(dict.fromkeys([x for x in labels if x == int(x)])))
    class_index = label_classes.index(label_target_class)
    # Create empty SHAP value matrix
    n_samples, n_features = X.shape
    n_selection_rounds = len(is_selected)
    mc_shap_values = np.zeros(shape=(n_samples, n_features, n_rounds, n_selection_rounds))
    # Compute SHAP values
    list_expected_value = []
    if verbose:
        ut.print_start_progress(start_message=f"ShapExplainer starts Monte Carlo estimation of SHAP values over {n_rounds} rounds.")
    for i in range(n_rounds):
        for j, selected_features in enumerate(is_selected):
            if verbose:
                pct_progress = j / len(is_selected)
                add_new_line = explainer_class in LIST_VERBOSE_SHAP_EXPLAINERS
                ut.print_progress(i=i+pct_progress, n=n_rounds, add_new_line=add_new_line)
            # Adjust fuzzy labels (labels between 0 and 1, e.g., 0.5 -> 50% 1 and 50% 0)
            if fuzzy_labeling:
                threshold = (j * (i + 1)) / (n_rounds * n_selection_rounds)
                labels_ = [x if x in [0, 1] else int(x >= threshold) for x in labels]
            else:
                labels_ = labels
            X_selected = X[:, selected_features]
            args = dict(list_model_classes=list_model_classes, list_model_kwargs=list_model_kwargs,
                        explainer_class=explainer_class, explainer_kwargs=explainer_kwargs,
                        class_index=class_index, n_background_data=n_background_data)
            _shap_values, _exp_val = _aggregate_shap_values(X_selected, labels=labels_, **args)
            mc_shap_values[:, selected_features, i, j] = _shap_values
            list_expected_value.append(_exp_val)
    # Averaging over rounds and selections
    if verbose:
        ut.print_end_progress(end_message=f"ShapExplainer finished Monte Carlo estimation and saved results.")
    shap_values = np.mean(mc_shap_values, axis=(2, 3))
    exp_val = np.mean(list_expected_value)
    return shap_values, exp_val



"""
This is a script for the backend of the dPULearn.compare_sets_negatives() method.
"""
import pandas as pd
import numpy as np


# I Helper Functions


# II Main Functions
def compare_sets_negatives_(list_labels=None, names=None, df_seq=None,
                            return_upset_data=None, remove_non_neg=True):
    """Create DataFrame for comparing sets of identified negatives"""
    combined_negatives = set()
    for labels in list_labels:
        negatives = set(np.where(labels == 0)[0])
        combined_negatives.update(negatives)
    data = {}
    reverse_cols = names is not None
    names = [f'Set {i+1}' for i, l in enumerate(list_labels)] if names is None else names
    for name, labels in zip(names, list_labels):
        data[name] = [(index in combined_negatives and labels[index] == 0) for index in range(len(labels))]
    df = pd.DataFrame(data)
    # Filter samples that are never identified as negatives
    mask = np.asarray(df.sum(axis=1) != 0)
    if not return_upset_data:
        if df_seq is not None:
            df = pd.concat([df_seq, df], axis=1)
        if remove_non_neg:
            df = df[mask]
        df_neg_comp = df
        return df_neg_comp
    # Create UpsetPlot data
    if reverse_cols:
        df = df[list(reversed(list(df)))]
    if remove_non_neg:
        df = df[mask]
    upset_data = df.groupby(list(df.columns)).size()
    return upset_data





"""
This is a script for the backend of the CPPPlot.ranking() method.
"""
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib as mpl

import aaanalysis.utils as ut
from .utils_feature import get_positions_


# I Helper Functions
# Adjust df_feat
def _adjust_df_feat(df_feat=None, col_dif=None):
    """Adjusts feature values in `df_feat` based on percentage scaling and sets the limits for difference columns."""
    df_feat = df_feat.copy()
    if max(df_feat[col_dif]) - min(df_feat[col_dif]) <= 2:
        df_feat[col_dif] *= 100
    return df_feat


# 1. Subplot: Feature position
def _add_pos_bars(ax=None, x=None, y=None, color="tab:gray", height=0.5, width=1.0, alpha=1.0):
    """Adds a rectangle (bar) to the axis `ax` at specified `x`, `y` position with given dimensions and style."""
    bar = mpl.patches.Rectangle((x, y), width=width, height=height, linewidth=0, color=color, zorder=4, clip_on=False,
                                alpha=alpha)
    ax.add_patch(bar)


def _add_position_bars(ax=None, df_feat=None):
    """Plots position bars for each feature in `df_feat` on the given axis `ax`."""
    for y, row in df_feat.iterrows():
        positions = [int(x) for x in row[ut.COL_POSITION].split(",")]
        is_segment = positions[-1] - positions[0] == len(positions) - 1
        # Add segment or pattern
        if is_segment:
            _add_pos_bars(ax=ax, x=positions[0] - 1, y=y - 0.25, width=len(positions))
        else:
            # Add pattern
            for pos in positions:
                _add_pos_bars(ax=ax, x=pos - 1, y=y - 0.25)

def _add_part_seq(ax=None, jmd_n_len=10, jmd_c_len=10, tmd_len=20, y=-0.75, height=0.2,
                 tmd_color="mediumspringgreen", jmd_color="blue", alpha=1.0, start=0.0):
    """Add colored box for sequence parts in figure"""
    list_color = [jmd_color, tmd_color, jmd_color]
    list_length = [jmd_n_len, tmd_len, jmd_c_len]
    # Add jmd_n
    for length, color in zip(list_length, list_color):
        bar = mpl.patches.Rectangle((start, y), width=length, height=height, linewidth=0, color=color, zorder=4,
                                    clip_on=False, alpha=alpha)
        start += length
        ax.add_patch(bar)


def _get_tmd_jmd_label(jmd_n_len=10, jmd_c_len=10, space=3):
    """Generates a label for TMD and JMD regions based on their lengths and spacing."""
    jmd_len = jmd_c_len + jmd_n_len
    name_tmd = ut.options["name_tmd"]
    name_jmd_n = ut.options["name_jmd_n"]
    name_jmd_c = ut.options["name_jmd_c"]
    # Space factors should be between 1 and max-1
    total_space = space*2
    space_n_factor = max(min(int(round(jmd_n_len/jmd_len*total_space)), total_space-1), 1)
    space_c_factor = total_space - space_n_factor
    x_label = ""
    x_label += name_jmd_n + " " * space_n_factor if jmd_n_len > 0 else " " * (4 + space_c_factor)
    x_label += name_tmd
    x_label += " " * space_c_factor + name_jmd_c if jmd_c_len > 0 else " " * (4 + space_n_factor)
    return x_label


def plot_feature_position(ax=None, df=None, n=20, space=3, tmd_len=20, jmd_n_len=10, jmd_c_len=10,
                          tmd_color="mediumspringgreen", jmd_color="blue", fontsize_label=None, tmd_jmd_alpha=0.075):
    """Plots the feature positions for a given DataFrame `df` on the axis `ax` with specified formatting parameters."""
    fig_height = plt.gcf().get_size_inches()[1]
    height = min([0.01 * n, 0.2]) * 5/fig_height
    plt.sca(ax)
    # Set y ticks
    plt.tick_params(axis='y', which='both', bottom=False, top=False, labelbottom=False)
    list_ticks = list(reversed(df[ut.COL_SUBCAT]))
    plt.yticks(range(0, n), list_ticks)
    # Set x ticks
    x_max = jmd_n_len + tmd_len + jmd_c_len
    plt.xticks(range(0, x_max))
    plt.tick_params(axis='x', which='both', bottom=False, top=False, labelbottom=False)
    sns.despine(top=True, right=True, left=False, bottom=True)
    args_len = dict(tmd_len=tmd_len, jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len)
    args_color = dict(tmd_color=tmd_color, jmd_color=jmd_color)
    # Add sequence part under plot
    _add_part_seq(ax=ax, y=len(df)-0.5, height=height, **args_len, **args_color)
    # Add sequence area in plot
    _add_part_seq(ax=ax, y=-0.5, height=len(df), alpha=tmd_jmd_alpha, **args_len, **args_color)
    _add_position_bars(ax=ax, df_feat=df)
    # Adjust xlabel
    x_label = _get_tmd_jmd_label(jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len, space=space)
    plt.xlabel(x_label, size=fontsize_label)


# 2. Subplot: Feature mean difference
def _add_annotation_extreme_val(sub_fig=None, max_neg_val=-2, min_pos_val=2, text_size=9):
    """Adds annotations for extreme values in a subplot `sub_fig`, defining thresholds for maximum negative and minimum positive values."""
    args = dict(va='center_baseline', size=text_size, xytext=(0.5, 0), textcoords='offset points',
                color="white")
    for p in sub_fig.patches:
        val = round(p.get_width(), 1)
        if val <= max_neg_val:
            x = max_neg_val + 1
            ha = "left"
            sub_fig.annotate(f"{int(val)}%", (x, p.get_y() + p.get_height() / 2), ha=ha, **args)
        if val >= min_pos_val:
            x = min_pos_val - 1
            ha = "right"
            sub_fig.annotate(f"{int(val)}%", (x, p.get_y() + p.get_height()/2), ha=ha, **args)


def plot_feature_mean_dif(ax=None, df=None, col_dif=None, n=20, xlim=(-22, 22), fontsize_annotation=8):
    """Plots the mean difference of features in `df` on the axis `ax`, with custom range `xlim` and annotation font size."""
    plt.sca(ax)
    colors = [ut.get_color_dif(mean_dif=x) for x in df[col_dif]]
    df["hue"] = colors     # Adjust to use palette for sns.barplot after v0.14.0
    args = dict(hue="hue", palette={x: x for x in colors}, legend=False)
    sub_fig = sns.barplot(ax=ax, data=df, y="feature", x=col_dif, **args)
    sns.despine(top=True, right=True, left=False, bottom=False)
    _add_annotation_extreme_val(sub_fig=sub_fig, text_size=fontsize_annotation,
                                max_neg_val=xlim[0], min_pos_val=xlim[1])
    # Add values for importance
    plt.axvline(x=0, color='gray', linestyle='-')
    plt.yticks(range(0, n), list(df[ut.COL_SUBCAT]))
    plt.xlim(xlim)
    x_tick_max = int(xlim[1]/5)*5 if not (0 < xlim[1] < 1) else int(xlim[1]*100/5)*5/100
    x_tick_min = int(xlim[0]/5)*5 if not (0 < abs(xlim[0]) < 1) else int(xlim[0]*100/5)*5/100
    plt.xticks(ticks=[x_tick_min, 0, x_tick_max], labels=[str(x_tick_min), "0", str(x_tick_max)])
    plt.ylabel("")


# 3. Subplot: Ranking based on absolute AUC, feature importance or feature impact
def _add_annotation_right(sub_fig=None, an_in_val=2, max_val=10.0, text_size=8):
    """Adds right-aligned annotations for feature importance in a subplot `sub_fig`, considering an inner value threshold and maximum value."""
    args = dict(va='center_baseline', size=text_size, xytext=(0.5, 0), textcoords='offset points')
    args_left = dict(ha='left', **args)
    args_right = dict(ha="right", **args, color="white")
    for p in sub_fig.patches:
        val = round(p.get_width(), 1)
        if val < an_in_val:
            sub_fig.annotate(f"{round(val, 1)}%", (val, p.get_y() + p.get_height()/2), **args_left)
        else:
            x = max_val if val > max_val else val
            sub_fig.annotate(f"{round(val, 1)}%", (x, p.get_y() + p.get_height()/2), **args_right)


def plot_feature_rank(ax=None, df_feat=None, n=20, xlim=(0, 4),
                      fontsize_annotation=8, col_imp=ut.COL_FEAT_IMPORT,
                      shap_plot=False, rank_info_xy=None):
    """Plots the feature ranking based on `df` on the axis `ax`, adjusting for SHAP values if `shap_plot` is True."""
    df_feat = df_feat.copy()
    plt.sca(ax)
    if shap_plot:
        colors = [ut.COLOR_SHAP_NEG if v < 0 else ut.COLOR_SHAP_POS for v in df_feat[col_imp]]
    else:
        colors = ["tab:gray"] * len(df_feat)
    df_feat[col_imp] = abs(df_feat[col_imp]).round(1)
    df_feat["hue"] = colors     # Adjust to use palette for sns.barplot after v0.14.0
    args = dict(hue="hue", palette={x: x for x in colors}, legend=False)
    sub_fig = sns.barplot(ax=ax, data=df_feat, y="feature", x=col_imp, **args)
    plt.yticks(range(0, n), list(df_feat[ut.COL_SUBCAT]))
    plt.ylabel("")

    # Set x-axis limits
    xlim = plt.xlim() if xlim is None else xlim
    x_max = max(df_feat[col_imp].max(), xlim[1])
    xlim = (xlim[0], x_max)
    plt.xlim(xlim)

    _add_annotation_right(sub_fig=sub_fig, text_size=fontsize_annotation, an_in_val=x_max/2, max_val=xlim[1])
    # Add legend
    str_sum = f"Σ={round(df_feat[col_imp].sum(), 1)}%"
    args = dict(ha="right", size=fontsize_annotation)
    rank_info_xy_default = (xlim[1]*1.1, n-2.5)
    _rank_info_xy = ut.adjust_tuple_elements(tuple_in=rank_info_xy,
                                            tuple_default=rank_info_xy_default)
    x, y = _rank_info_xy
    plt.text(x, y, str_sum, weight="normal", **args)
    if shap_plot:
        plt.text(x, y+1, "negative", weight="bold", color=ut.COLOR_SHAP_NEG, va="top", **args)
        plt.text(x, y+1, "positive", weight="bold", color=ut.COLOR_SHAP_POS, va="bottom", **args)


# II Main Functions
def plot_ranking(df_feat=None,
                 n_top=15, rank=False,
                 col_dif=None, col_imp=None,
                 shap_plot=False,
                 figsize=(7, 5),
                 tmd_len=20, jmd_n_len=10, jmd_c_len=10,
                 tmd_color="mediumspringgreen", jmd_color="blue",
                 tmd_jmd_alpha=0.075,
                 name_test="test", name_ref="ref",
                 fontsize_titles=11,
                 fontsize_labels=11,
                 fontsize_annotations=11,
                 tmd_jmd_space=2,
                 xlim_dif=(-17.5, 17.5),
                 xlim_rank=(0, 4),
                 rank_info_xy=None):
    """Plot ranking of feature DataFrame"""
    # Adjust df_feat
    if rank:
        df_feat = df_feat.sort_values(by=col_imp, ascending=False)
    df_feat = df_feat.head(n_top).copy().reset_index(drop=True)
    df_feat = _adjust_df_feat(df_feat=df_feat, col_dif=col_dif)
    df_feat[ut.COL_POSITION] = get_positions_(features=df_feat[ut.COL_FEATURE],
                                              tmd_len=tmd_len, jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len)
    # Plotting (three subplots)
    fig, axes = plt.subplots(1, 3, sharey=True, figsize=figsize)
    # 1. Plot feature positions
    plot_feature_position(ax=axes[0], df=df_feat, n=n_top, space=tmd_jmd_space,
                          tmd_len=tmd_len, jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len,
                          tmd_color=tmd_color, jmd_color=jmd_color, tmd_jmd_alpha=tmd_jmd_alpha,
                          fontsize_label=fontsize_labels)
    plt.title(ut.LABEL_FEAT_POS, x=0, weight="bold", fontsize=fontsize_titles)
    # 2. Barplot mean difference
    plot_feature_mean_dif(ax=axes[1], df=df_feat,
                          n=n_top, col_dif=col_dif, xlim=xlim_dif,
                          fontsize_annotation=fontsize_annotations)
    sns.despine(ax=axes[1], top=True, right=True, left=True, bottom=False)
    axes[1].set_title(f"Mean difference\nof feature value",
                      size=fontsize_titles, weight="bold")
    label_mean_dif = f"{name_test} - {name_ref} [%]"
    axes[1].set_xlabel(label_mean_dif, size=fontsize_labels)
    # 3. Barplot importance
    plot_feature_rank(ax=axes[2], df_feat=df_feat, n=n_top, xlim=xlim_rank,
                      col_imp=col_imp, shap_plot=shap_plot, rank_info_xy=rank_info_xy,
                      fontsize_annotation=fontsize_annotations)
    axes[2].set_title(f"{ut.LABEL_FEAT_RANKING}\n(top {n_top} features)",
                      size=fontsize_titles, ha="center", weight="bold")
    label_ranking = ut.LABEL_FEAT_IMPACT if shap_plot else ut.LABEL_FEAT_IMPORT
    axes[2].set_xlabel(label_ranking, size=fontsize_labels)
    # Adjust axis
    for i, ax in enumerate(axes):
        ax.tick_params(which='major', axis="both", labelsize=fontsize_labels)
        if i > 0:
            ax.tick_params(which='major', axis="y", length=0, labelsize=0)
    return fig, axes


"""
This is a script for the backend of the SequenceFeature() object,
a supportive class for the CPP feature engineering.
"""
import pandas as pd
import numpy as np

import aaanalysis.utils as ut
from .utils_feature import get_positions_, get_feature_matrix_, get_amino_acids_, add_scale_info_
from ._split import SplitRange
from ._utils_feature_stat import add_stat_


# I Helper Functions


# II Main functions
# Parts and splits
def get_split_kws_(n_split_min=1, n_split_max=15, steps_pattern=None, n_min=2, n_max=4, len_max=15,
                   steps_periodicpattern=None, split_types=None):
    """Get split kws for CPP class"""
    if split_types is None:
        split_types = ut.LIST_SPLIT_TYPES
    if steps_pattern is None:
        # Differences between interacting amino acids in helix (without gaps) include 6, 7 ,8 to include gaps
        steps_pattern = [3, 4]
    if steps_periodicpattern is None:
        steps_periodicpattern = [3, 4]  # Differences between interacting amino acids in helix (without gaps)
    split_kws = {ut.STR_SEGMENT: dict(n_split_min=n_split_min, n_split_max=n_split_max),
                 ut.STR_PATTERN: dict(steps=steps_pattern, n_min=n_min, n_max=n_max, len_max=len_max),
                 ut.STR_PERIODIC_PATTERN: dict(steps=steps_periodicpattern)}
    split_kws = {x: split_kws[x] for x in split_types}
    return split_kws


# Features
def get_features_(list_parts=None, split_kws=None, list_scales=None):
    """Create list of all feature ids for given Parts, Splits, and Scales"""
    spr = SplitRange()
    features = []
    for split_type in split_kws:
        args = split_kws[split_type]
        labels_s = getattr(spr, "labels_" + split_type.lower())(**args)
        features.extend(["{}-{}-{}".format(p.upper(), s, sc) for p in list_parts for s in labels_s
                         for sc in list_scales])
    return features


def get_feature_names_(features=None, df_cat=None, tmd_len=20, jmd_c_len=10, jmd_n_len=10, start=1):
    """Convert feature ids (PART-SPLIT-SCALE) into feature names (scale name [positions])."""
    feat_positions = get_positions_(features=features, tmd_len=tmd_len, jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len,
                                    start=start)
    dict_scales = dict(zip(df_cat[ut.COL_SCALE_ID], df_cat[ut.COL_SUBCAT]))
    feat_names = []
    for feat_id, pos in zip(features, feat_positions):
        part, split, scale = feat_id.split("-")
        split_type = split.split("(")[0]
        if split_type == ut.STR_SEGMENT and len(pos.split(",")) > 2:
            pos = pos.split(",")[0] + "-" + pos.split(",")[-1]
        if split_type == ut.STR_PERIODIC_PATTERN:
            step = split.split("+")[1].split(",")[0]
            pos = pos.split(",")[0] + ".." + step + ".." + pos.split(",")[-1]
        feat_names.append(f"{dict_scales[scale]} [{pos}]")
    return feat_names


def get_df_feat_(features=None, df_parts=None, labels=None,
                 label_test=1, label_ref=0, df_scales=None, df_cat=None,
                 accept_gaps=False, parametric=False,
                 start=1, tmd_len=20, jmd_c_len=10, jmd_n_len=10, n_jobs=1):
    """Create df feature for comparing groups and or samples"""
    X = get_feature_matrix_(features=features, df_parts=df_parts, df_scales=df_scales,
                            accept_gaps=accept_gaps, n_jobs=n_jobs)
    mask_test = [x == label_test for x in labels]
    mask_ref = [x == label_ref for x in labels]
    mean_dif = X[mask_test].mean(axis=0) - X[mask_ref].mean(axis=0)
    abs_mean_dif = abs(mean_dif)
    std_test = X[mask_test].std(axis=0)
    df = pd.DataFrame(zip(features, abs_mean_dif, std_test),
                      columns=[ut.COL_FEATURE, ut.COL_ABS_MEAN_DIF, ut.COL_STD_TEST])
    df = add_stat_(df=df, X=X, labels=labels, parametric=parametric, label_test=label_test, label_ref=label_ref)
    df = add_scale_info_(df_feat=df, df_cat=df_cat)
    df[ut.COL_POSITION] = get_positions_(features=features, start=start, jmd_n_len=jmd_n_len, tmd_len=tmd_len,
                                         jmd_c_len=jmd_c_len)
    if sum(mask_test) == 1:
        position = np.where(labels == label_test)[0][0]
        jmd_n_seq, tmd_seq, jmd_c_seq = df_parts[[ut.COL_JMD_N, ut.COL_TMD, ut.COL_JMD_C]].iloc[position].values
        df[ut.COL_AA_TEST] = get_amino_acids_(features=features, jmd_n_seq=jmd_n_seq, tmd_seq=tmd_seq,
                                              jmd_c_seq=jmd_c_seq)
    if sum(mask_ref) == 1:
        position = np.where(labels == label_ref)[0][0]
        jmd_n_seq, tmd_seq, jmd_c_seq = df_parts[[ut.COL_JMD_N, ut.COL_TMD, ut.COL_JMD_C]].iloc[position].values
        df[ut.COL_AA_REF] = get_amino_acids_(features=features, jmd_n_seq=jmd_n_seq, tmd_seq=tmd_seq,
                                             jmd_c_seq=jmd_c_seq)
    return df



"""
This is a script for internal plotting utility functions used in the backend.
"""
import seaborn as sns
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import matplotlib.ticker as mticker


# Helper functions
def _get_color_map(labels=None, colors=None, sort_unique_labels=True):
    """Get colormap"""
    if sort_unique_labels:
        unique_labels = sorted(set(labels))
    else:
        unique_labels = labels
    if isinstance(colors, list) and len(colors) != 1:
        if len(colors) < len(unique_labels):
            colors *= len(unique_labels)
        color_map = {label: colors[i] for i, label in enumerate(unique_labels)}
    else:
        colors = colors[0] if isinstance(colors, list) else colors
        color_map = {label: colors for label in unique_labels}
    return color_map, unique_labels


def _get_positions_lengths_colors(labels, color_map):
    """Get the positions """
    positions, lengths, colors = [], [], []
    current_label, start_pos = labels[0], 0
    for i, label in enumerate(labels + [None]):
        if label != current_label or i == len(labels):
            positions.append(start_pos)
            length = i - start_pos
            lengths.append(length)
            colors.append(color_map[current_label])
            start_pos = i
            current_label = label
    return positions, lengths, colors


def _get_xy_wh(ax=None, position=None, pos=None, bar_spacing=None, length=None, bar_width=None):
    """Get x and y position together with width and height/length"""
    x_min, x_max = ax.get_xlim()
    y_min, y_max = ax.get_ylim()
    if position == 'bottom':
        y = y_min + bar_spacing
        return (pos, y), (length, bar_width)
    elif position == 'top':
        y = y_max - bar_spacing - bar_width
        return (pos, y), (length, bar_width)
    elif position == 'left':
        x =  x_min - bar_spacing - bar_width
        return (x, pos), (bar_width, length)
    elif position == 'right':
        x = x_max + bar_spacing
        return (x, pos), (bar_width, length)
    else:
        raise ValueError("Position should be 'left', 'right', 'top', or 'bottom'.")


def _get_xy_hava(position=None, xy=None, wh=None, label_spacing_factor=1.5):
    """Get x and y position together with horizontal alignment and vertical alignment"""
    bar_width = wh[1] if position in ['bottom', 'top'] else wh[0]
    if position == 'bottom':
        text_x = xy[0] + wh[0] / 2
        text_y = xy[1] + bar_width * label_spacing_factor
        ha, va = 'center', 'top'
    elif position == 'top':
        text_x = xy[0] + wh[0] / 2
        text_y = xy[1] + wh[1] - bar_width * (label_spacing_factor-0.25)
        ha, va = 'center', 'bottom'
    elif position == 'left':
        text_x = xy[0] - bar_width * (label_spacing_factor-1)
        text_y = xy[1] + wh[1] / 2
        ha, va = 'right', 'center'
    else:  # Assuming position == 'right'
        text_x = xy[0] + wh[0] + bar_width * (label_spacing_factor-1)
        text_y = xy[1] + wh[1] / 2
        ha, va = 'left', 'center'
    return text_x, text_y, ha, va


def _add_bar_labels(ax=None, bar_labels_align=None, position=None, bar_width=None,
                    labels=None, positions=None, lengths=None, bar_labels=None, bar_spacing=None,
                    label_spacing_factor=1.5):
    """Add labels next to the bars"""
    label_map = {label: bar_labels[i] for i, label in enumerate(sorted(set(labels)))}
    rotation = 0 if bar_labels_align == 'horizontal' else 90
    for pos, length in zip(positions, lengths):
        xy, wh = _get_xy_wh(ax=ax, position=position, pos=pos, bar_spacing=bar_spacing, length=length, bar_width=bar_width)
        text_x, text_y, ha, va = _get_xy_hava(position=position, xy=xy, wh=wh, label_spacing_factor=label_spacing_factor)
        ax.text(text_x, text_y, label_map[labels[int(pos + length / 2)]], ha=ha, va=va, rotation=rotation,
                transform=ax.transData, clip_on=False)


def _add_text_labels(ax=None, position=None, bar_width=None, bar_spacing=None, label_spacing_factor=1.5,
                     nx=None, ny=None, xtick_label_rotation=45, ytick_label_rotation=0):
    """Add text labels next to the bars."""
    # Obtain tick labels and locations based on the specified position
    if position in ['left', 'right']:
        tick_labels = [item.get_text() for item in ax.get_yticklabels()]
        tick_locs = ax.get_yticks()
        ax.yaxis.set_visible(False)  # Hide the original y-axis
    else:  # ['top', 'bottom']
        tick_labels = [item.get_text() for item in ax.get_xticklabels()]
        tick_locs = ax.get_xticks()
        ax.xaxis.set_visible(False)  # Hide the original x-axis
    label_spacing = bar_spacing * label_spacing_factor
    for idx, label in enumerate(tick_labels):
        if position == 'left':
            ax.text(-label_spacing - bar_width, tick_locs[idx], label,
                    ha='right', va='center', rotation=ytick_label_rotation)
        elif position == 'right':
            ax.text(nx + label_spacing + bar_width, tick_locs[idx], label,
                    ha='left', va='center', rotation=ytick_label_rotation)
        elif position == 'top':
            ax.text(tick_locs[idx], - label_spacing - bar_width, label,
                    ha='left', va='bottom', rotation=xtick_label_rotation)
        elif position == 'bottom':
            ax.text(tick_locs[idx], ny + label_spacing + bar_width, label,
                    ha='right', va='top', rotation=xtick_label_rotation)


# Main function
def plot_add_bars(ax=None, labels=None, colors='tab:gray', bar_position='left',
                  bar_spacing=0.05,  bar_width=0.1, bar_labels=None, label_spacing_factor=1.5,
                  bar_labels_align='horizontal', set_tick_labels=True,
                  xtick_label_rotation=45, ytick_label_rotation=0):
    """Add colored bars along a specified axis of the plot based on label grouping."""
    if labels is None:
        raise ValueError("'labels' should not be None")
    # Get the number of plotted items
    nx, ny = int(max(ax.get_xlim())), int(max(ax.get_ylim()))
    num_plotted_items = ny if bar_position in ['left', 'right'] else nx
    if not isinstance(labels, list) and labels is not None:
        labels = list(labels)
    # Check if labels match the shape of the data
    if len(labels) != num_plotted_items:
        raise ValueError(f"Mismatch: The number of labels ({len(labels)}) must match to number of plotted items ({num_plotted_items}).")
    single_color = isinstance(colors, str) or (isinstance(colors, (list, tuple)) and len(set(colors)) == 1)
    sort_unique_labels = len(labels) != len(colors)
    color_map, _ = _get_color_map(labels=labels, colors=colors, sort_unique_labels=sort_unique_labels)
    positions, lengths, colors = _get_positions_lengths_colors(labels, color_map)
    args_pos = dict(position=bar_position, bar_width=bar_width, bar_spacing=bar_spacing)
    if bar_labels is not None:
        _add_bar_labels(ax=ax, bar_labels_align=bar_labels_align,
                        labels=labels, positions=positions, lengths=lengths,
                        bar_labels=bar_labels, label_spacing_factor=label_spacing_factor, **args_pos)
    # Adding bars
    args = dict(transform=ax.transData, clip_on=False)
    for pos, length, bar_color in zip(positions, lengths, colors):
        xy, wh = _get_xy_wh(ax=ax, pos=pos, length=length, **args_pos)
        # Add edgecolor if only one color is specified
        edgecolor = "white" if single_color else bar_color
        ax.add_patch(mpatches.Rectangle(xy=xy, width=wh[0], height=wh[1],
                                        facecolor=bar_color, edgecolor=edgecolor, linewidth=0.5,
                                        **args))
    if set_tick_labels:
        _add_text_labels(ax=ax, **args_pos, label_spacing_factor=label_spacing_factor,
                         xtick_label_rotation=xtick_label_rotation, ytick_label_rotation=ytick_label_rotation,
                         nx=nx, ny=ny)


def adjust_spine_to_middle(ax=None):
    """Adjust spines to be in middle if data range from <0 to >0"""
    min_val, max_val = ax.get_xlim()
    if max_val > 0 and min_val >= 0:
        sns.despine(ax=ax)
    else:
        sns.despine(ax=ax, left=True)
        current_lw = ax.spines['bottom'].get_linewidth()
        ax.axvline(0, color='black', linewidth=current_lw)
        val = max([abs(min_val), abs(max_val)])
        ax.set_xlim(-val, val)
    return ax


def x_ticks_0(ax):
    """Apply custom formatting for x-axis ticks."""
    def custom_x_ticks(x, pos):
        """Format x-axis ticks."""
        if x % 1 == 0:  # Check if number is an integer
            return f'{int(x)}'  # Format as integer
        else:
            return f'{x:.2f}'  # Format as float with two decimal places
    ax.xaxis.set_major_formatter(mticker.FuncFormatter(custom_x_ticks))


def adjust_tuple_elements(tuple_in=None, tuple_default=None):
    """Replace elements of input tuple that are None with the corresponding element from default tuple."""
    if tuple_in is not None:
        tuple_out = tuple(i if i is not None else def_i for i, def_i in zip(tuple_in,  tuple_default))
    else:
        tuple_out =  tuple_default
    return tuple_out


"""
This is a script for setting system level options for AAanalysis.
"""
from typing import Dict, Any
import os

from ._utils.check_type import check_bool, check_number_val, check_number_range, check_str
from ._utils.check_data import check_df

# System level options
_dict_options = {
    'verbose': "off",
    'random_state': "off",
    'allow_multiprocessing': True,
    'name_tmd': "TMD",
    'name_jmd_n': "JMD-N",
    'name_jmd_c': "JMD-C",
    'ext_len': 0,
    'df_scales': None,
    'df_cat': None,
}

# Check system level (option) parameters or depending parameters
def check_verbose(verbose=None):
    """Check if general verbosity is on or off. Adjusted based on options setting and value provided to object"""
    global_verbose = options["verbose"]
    if global_verbose != "off":
        # System level verbosity
        check_bool(name="verbose (option)", val=global_verbose)
        verbose = global_verbose
    else:
        check_bool(name="verbose", val=verbose)
    return verbose


def check_random_state(random_state=None):
    """Adjust random state if global is not 'off' (default)"""
    global_random_state = options["random_state"]
    args = dict(min_val=0, accept_none=True, just_int=True)
    if global_random_state != "off":
        # System-level random state
        check_number_range(name="random_state (option)", val=global_random_state, **args)
        random_state = global_random_state
    else:
        check_number_range(name="random_state", val=random_state, **args)
    return random_state


def check_n_jobs(n_jobs=None):
    """Adjust n_jobs to 1 if multiprocessing is not allowed"""
    allow_multiprocessing = options["allow_multiprocessing"]
    check_bool(name="allow_multiprocessing (options)", val=allow_multiprocessing)
    # Disable multiprocessing
    if not allow_multiprocessing:
        n_jobs = 1
        os.environ['LOKY_MAX_CPU_COUNT'] = "1"
    # Set n_jobs to maximum number of CPUs
    if n_jobs == -1:
        n_jobs = os.cpu_count()
    # Check which n_jobs are allowed
    check_number_val(name="j_jobs", val=n_jobs, accept_none=True)
    if n_jobs is None or n_jobs >= 1:
        check_number_range(name="n_jobs", val=n_jobs, accept_none=True, just_int=True, min_val=1)
    return n_jobs


# DEV: Parameters are used as directive to get better documentation style
# Enables setting of system level variables like in matplotlib
def _check_option(name_option="", option=None):
    """Check if option is valid"""
    if name_option == "verbose":
        if option != "off":
            check_verbose(verbose=option)
    if name_option == "random_state":
        if option != "off":
            check_random_state(random_state=option)
    if name_option == "allow_multiprocessing":
        check_bool(name=name_option, val=option)
    if "name" in name_option:
        check_str(name=name_option, val=option, accept_none=False)
    if name_option == "ext_len":
        check_number_range(name=name_option, val=option, min_val=0, accept_none=False)
    if "df" in name_option:
        check_df(name=name_option, df=option, accept_none=False)


class Settings:
    """
     A class for managing system-level settings for AAanalysis.

    This class mimics a dictionary-like interface, allowing the setting and retrieving
    of system-level options. It is designed to be used as a single global instance, ``options``.

    Parameters
    ----------
    The following options can be set:

    verbose : bool or 'off', default='off'
        Sets verbose mode to ``True`` or ``False`` globally if not 'off'.
    random_state : int, None, or 'off', default='off'
        The seed used by the random number generator.

        * If set to a positive integer, results of stochastic processes are consistent, enabling reproducibility.
        * If set to ``None``, stochastic processes will be truly random.
        * If set to 'off', no global random state variable will be set, allowing the underlying libraries to use
          their default random state behavior.

    allow_multiprocessing : bool, default=True
        Whether multiprocessing is allowed in general. If ``False``, ``n_jobs`` is automatically set to 1.
    name_tmd : str, default='TMD'
        Name of target middle domain (TMD) used in CPP plots.
    name_jmd_n : str, default='JMD_N'
        Name of N-terminal juxta middle domain (JMD-N) used in CPP plots.
    name_jmd_c : str, default='JMD_C'
        Name of C-terminal juxta middle domain (JMD-C) used in CPP plots.
    ext_len : int, default=0
        Length of TMD-extending part (starting from C and N terminal part of TMD, >=0). Disabled (set to 0) by default.
    df_scales : DataFrame, optional
        Scale DataFrame used in CPP algorithm. Adjust on system level if non-default scales are used.
        If ``None``, AAanalysis framework will use the scale DataFrame loaded by :func:`load_scales` with ``name='scales'``.
    df_cat : DataFrame, optional
        Scale category DataFrame used in CPP algorithm. Adjust on system level if non-default scale categories are used.
        If ``None``, AAanalysis framework will use the scale category DataFrame loaded by :func:`load_scales` with ``name='scales_cat'``.

    See Also
    --------
    * :class:`numpy.random.RandomState` for details on the ``random_state`` variable used to make stochastic processes
      yielding consistent results.
    * :class:`SequenceFeature` for definition of sequence ``Parts``.
    * :func:`load_scales` for details on scale and scale category DataFrames.

    Warnings
    --------
    * Multiprocessing Compatibility: Enabling multiprocessing (``allow_multiprocessing=True``)
      can lead to issues in environments that don't support forking or when interfacing with
      certain libraries. If encountering errors, consider setting ``allow_multiprocessing=False``.
      Note that this may affect performance in computation-intensive operations.

    Examples
    --------
    .. include:: examples/options.rst
    """
    def __init__(self):
        self._settings: Dict[str, Any] = _dict_options.copy()

    def __getitem__(self, key: str) -> Any:
        """Retrieve a setting's value using dict-like access."""
        return self._settings.get(key, None)

    def __setitem__(self, key: str, value: Any) -> None:
        """Set a setting's value using dict-like access.
           Prevent adding new keys that are not already in the system options."""
        if key in self._settings:
            _check_option(name_option=key, option=value)
            self._settings[key] = value
        else:
            valid_options = list(_dict_options.keys())
            raise KeyError(f"'{key}' is not valid options. Valid options are: {valid_options}.")

    def __contains__(self, key: str) -> bool:
        """Check if a key is in the settings."""
        return key in self._settings

    def __str__(self) -> str:
        """Return a string representation of the settings dictionary."""
        return str(self._settings)

# Global settings instance
options = Settings()



"""
This is a script for the backend of the AAclust.eval method.
"""
import pandas as pd
import numpy as np

from sklearn.metrics import silhouette_score, calinski_harabasz_score
from aaanalysis.utils import bic_score_

import aaanalysis.utils as ut


# I Helper Functions
def _evaluate_clustering(X, labels=None):
    """Evaluate clustering results using BIC, CH, SC scores"""
    # Bayesian Information Criterion
    bic = bic_score_(X, labels)
    # Calinski-Harabasz Index
    ch = calinski_harabasz_score(X, labels)
    # Silhouette Coefficient
    sc = silhouette_score(X, labels)
    return bic, ch, sc

# II Main function
@ut.catch_runtime_warnings()
def evaluate_clustering(X, list_labels=None, names_datasets=None):
    """Evaluate sets of clustering labels"""
    list_evals = []
    warn_ch = False
    warn_sc = False
    for labels in list_labels:
        n_clusters = len(set(labels))
        bic, ch, sc = _evaluate_clustering(X, labels=labels)
        if np.isnan(ch):
            ch = 0
            warn_ch = True
        if np.isnan(sc):
            sc = -1
            warn_sc = True
        list_evals.append([n_clusters, bic, ch, sc])
    # Create the DataFrame
    df_eval = pd.DataFrame(list_evals, columns=ut.COLS_EVAL_AACLUST)
    df_eval = ut.add_names_to_df_eval(df_eval=df_eval, names=names_datasets)
    return df_eval, warn_ch, warn_sc




"""
Basic utility check functions for type checking
"""
import pandas as pd
import numpy as np

from ._utils import add_str
from .utils_types import VALID_INT_TYPES, VALID_INT_FLOAT_TYPES


# Type checking functions
def check_number_val(name=None, val=None, accept_none=False, just_int=False, str_add=None):
    """Check if value is a valid integer or float"""
    if val is None:
        if not accept_none:
            str_error = add_str(str_error=f"'{name}' should not be None.", str_add=str_add)
            raise ValueError(str_error)
        return None
    if just_int is None:
        raise ValueError("'just_int' must be specified")
    # Define valid types for integers and floating points
    valid_types = VALID_INT_TYPES if just_int else VALID_INT_TYPES + VALID_INT_FLOAT_TYPES
    type_description = "an integer" if just_int else "a float or an integer"
    if not isinstance(val, valid_types):
        str_error = add_str(str_error=f"'{name}' should be {type_description}, but got {type(val).__name__}.",
                            str_add=str_add)
        raise ValueError(str_error)


def check_number_range(name=None, val=None, min_val=0, max_val=None, exclusive_limits=False,
                       accept_none=False, just_int=None, str_add=None):
    """Check if value of given name is within defined range"""
    if val is None:
        if not accept_none:
            str_error = add_str(str_error=f"'{name}' should not be None.", str_add=str_add)
            raise ValueError(str_error)
        return None
    if just_int is None:
        raise ValueError("'just_int' must be specified")
    # Define valid types for integers and floating points
    valid_types = VALID_INT_TYPES if just_int else VALID_INT_TYPES + VALID_INT_FLOAT_TYPES
    # Verify the value's type and range
    type_description = "an integer" if just_int else "a float or an integer"
    if not isinstance(val, valid_types):
        str_error = add_str(str_error=f"'{name}' should be {type_description}, but got {type(val).__name__}.",
                            str_add=str_add)
        raise ValueError(str_error)
    # Min and max values are excluded from allowed values
    if exclusive_limits:
        if val <= min_val or (max_val is not None and val >= max_val):
            range_desc = f"n > {min_val}" if max_val is None else f"{min_val} < n < {max_val}"
            str_error = add_str(str_error=f"'{name}' should be {type_description} with {range_desc}, but got {val}.",
                                str_add=str_add)
            raise ValueError(str_error)
    else:
        if val < min_val or (max_val is not None and val > max_val):
            range_desc = f"n >= {min_val}" if max_val is None else f"{min_val} <= n <= {max_val}"
            str_error = add_str(str_error=f"'{name}' should be {type_description} with {range_desc}, but got {val}.",
                                str_add=str_add)
            raise ValueError(str_error)
    return val


def check_str(name=None, val=None, accept_none=False, return_empty_string=False, str_add=None):
    """Check type string"""
    if val is None:
        if not accept_none:
            str_error = add_str(str_error=f"'{name}' should not be None.", str_add=str_add)
            raise ValueError(str_error)
        return "" if return_empty_string else None
    if not isinstance(val, str):
        str_error = add_str(str_error=f"'{name}' ('{val}') should be string.",
                            str_add=str_add)
        raise ValueError(str_error)
    return val


def check_bool(name=None, val=None, accept_none=False, str_add=None):
    """Check if the provided value is a boolean."""
    if val is None:
        if not accept_none:
            str_error = add_str(str_error=f"'{name}' should not be None.", str_add=str_add)
            raise ValueError(str_error)
        return None
    if not isinstance(val, bool):
        str_error = add_str(str_error=f"'{name}' ({val}) should be bool.",
                            str_add=str_add)
        raise ValueError(str_error)


def check_dict(name=None, val=None, accept_none=False, str_add=None):
    """Check if the provided value is a dictionary."""
    if val is None:
        if not accept_none:
            str_error = add_str(str_error=f"'{name}' should not be None.", str_add=str_add)
            raise ValueError(str_error)
        return None
    if not isinstance(val, dict):
        str_error = add_str(str_error=f"'{name}' ({val}) should be a dictionary.",
                            str_add=str_add)
        raise ValueError(str_error)


def check_tuple(name=None, val=None, n=None, check_number=True, accept_none=False,
                accept_none_number=False, str_add=None):
    """Check if the provided value is a tuple, optionally of a certain length and containing only numbers."""
    if val is None:
        if not accept_none:
            str_error = add_str(str_error=f"'{name}' should not be None.", str_add=str_add)
            raise ValueError(str_error)
        return None
    if not isinstance(val, tuple):
        str_error = add_str(str_error=f"'{name}' ({val}) should be a tuple.",
                            str_add=str_add)
        raise ValueError(str_error)
    if n is not None and len(val) != n:
        str_error = add_str(str_error=f"'{name}' ({val}) should be a tuple with {n} elements.",
                            str_add=str_add)
        raise ValueError(str_error)
    if n is not None and check_number:
        for v in val:
            check_number_val(name=name, val=v, just_int=False, accept_none=accept_none_number,
                             str_add=str_add)


def check_list_like(name=None, val=None, accept_none=False, convert=True, accept_str=False, min_len=None,
                    check_all_non_neg_int=False, check_all_non_none=True, check_all_str_or_convertible=False,
                    str_add=None):
    """Check if the value is list-like, optionally converting it to a list, and performing additional checks."""
    if val is None:
        if not accept_none:
            str_error = add_str(str_error=f"'{name}' should not be None.", str_add=str_add)
            raise ValueError(str_error)
        return None
    if not convert:
        if not isinstance(val, list):
            str_error = add_str(str_error=f"'{name}' (type: {type(val)}) should be a list.",
                                str_add=str_add)
            raise ValueError(str_error)
    elif accept_str and isinstance(val, str):
        return [val]
    else:
        allowed_types = (list, tuple, np.ndarray, pd.Series)
        if not isinstance(val, allowed_types):
            str_error = add_str(str_error=f"'{name}' (type: {type(val)}) should be one of {allowed_types}.",
                                str_add=str_add)
            raise ValueError(str_error)
        if isinstance(val, np.ndarray) and val.ndim != 1:
            str_error = add_str(str_error=f"'{name}' is a multi-dimensional numpy array and cannot be considered as a list.",
                                str_add=str_add)
            raise ValueError(str_error)
        val = list(val) if isinstance(val, (np.ndarray, pd.Series)) else val
    if check_all_non_none:
        n_none = len([x for x in val if x is None])
        if n_none > 0:
            str_error = add_str(str_error=f"'{name}' should not contain 'None' (n={n_none})",
                                str_add=str_add)
            raise ValueError(str_error)
    if check_all_non_neg_int:
        if any(type(i) != int or i < 0 for i in val):
            str_error = add_str(str_error=f"'{name}' should only contain non-negative integers.",
                                str_add=str_add)
            raise ValueError(str_error)
    if check_all_str_or_convertible:
        wrong_elements = [x for x in val if not isinstance(x, (str, int, float, np.number))]
        if len(wrong_elements) > 0:
            str_error = add_str(str_error=f"The following elements in '{name}' are not strings or"
                                          f" reasonably convertible: {wrong_elements}",
                                str_add=str_add)
            raise ValueError(str_error)
        else:
            val = [str(x) for x in val]
    if min_len is not None and len(val) < min_len:
        str_error = add_str(str_error=f"'{name}' should not contain at least {min_len} elements",
                            str_add=str_add)
        raise ValueError(str_error)
    return val
